from abc import ABC, abstractmethod
from copy import deepcopy

from spuco.evaluate import Evaluator


class BaseRobustTrain(ABC):
    """
    Abstract base class for InvariantTrain methods
    Provides support for worst group accuracy early stopping
    """
    def __init__(
        self, 
        val_evaluator: Evaluator = None, 
        verbose: bool = False
    ):
        """
        Initializes the model trainer.

        :param val_evaluator: Evaluator object for validation evaluation. Default is None.
        :type val_evaluator: Evaluator, optional
        :param verbose: Whether to print training progress. Default is False.
        :type verbose: bool
        """
        self.val_evaluator = val_evaluator
        self._best_model = None 
        self._best_wg_acc = -1
        self.verbose = verbose
        self.trainer = None
        
    def train(self):
        """
        Train for specified number of epochs (and do early stopping if val_evaluator given)
        """
        for epoch in range(self.num_epochs):
            self.train_epoch(epoch)
            if self.val_evaluator is not None:
                self.val_evaluator.evaluate()
                if self.val_evaluator.worst_group_accuracy[1] > self._best_wg_acc:
                    self._best_wg_acc = self.val_evaluator.worst_group_accuracy[1]
                    self._best_model = deepcopy(self.trainer.model)
                    self._best_epoch = epoch
                if self.verbose:
                    print('Epoch {}: Val Worst-Group Accuracy: {}'.format(epoch, self.val_evaluator.worst_group_accuracy[1]))
                    print('Best Val Worst-Group Accuracy: {}'.format(self._best_wg_acc))
                
    def train_epoch(self, epoch: int):
        """
        Trains the model for a single epoch.

        :param epoch: The current epoch number.
        :type epoch: int
        """
        self.trainer.train_epoch(epoch)
        
    @property
    def best_model(self):
        """
        Property for accessing the best model.

        :return: The best model.
        :rtype: Any
        :raises NotImplementedError: If no val_evaluator is set to get worst group validation accuracy.
        """
        if self.val_evaluator is None:
            raise NotImplementedError("Cannot get best model if no val_evaluator set to \
                get worst group validation accuracy.")
        else:
            return self._best_model

    @property
    def best_wg_acc(self):
        """
        Property for accessing the best worst group validation accuracy.

        :return: The best worst group validation accuracy.
        :rtype: Any
        :raises NotImplementedError: If no val_evaluator is passed.
        """
        if self.val_evaluator is None:
            raise NotImplementedError("Cannot get worst group validation accuracy \
                no val_evaluator passed.")
        else:
            return self._best_wg_acc
        
    @property
    def best_epoch(self):
        """
        Property for accessing the best epoch number.

        :return: The best epoch number.
        :rtype: Any
        :raises NotImplementedError: If no val_evaluator is passed.
        """
        if self.val_evaluator is None:
            raise NotImplementedError("Cannot get early stopping epoch \
                no val_evaluator passed.")
        else:
            return self._best_epoch


import random

import numpy as np
import torch
from torch import nn

from spuco.utils.random_seed import seed_randomness


class SpuCoModel(nn.Module):
    """
    Wrapper module to allow for methods that use penultimate layer embeddings
    to easily access this via *backbone*
    """
    def __init__(
        self,
        backbone: nn.Module, 
        representation_dim: int,
        num_classes: int
    ):
        """
        Initializes a SpuCoModel 

        :param backbone: The backbone network.
        :type backbone: torch.nn.Module
        :param representation_dim: The dimensionality of the penultimate layer embeddings.
        :type representation_dim: int
        :param num_classes: The number of output classes.
        :type num_classes: int
        """
        
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)
        super().__init__()
        self.backbone = backbone 
        self.classifier = nn.Linear(representation_dim, num_classes)

    def forward(self, x):
        """
        Forward pass of the SpuCoModel.

        :param x: Input tensor.
        :type x: torch.Tensor
        :return: Output tensor.
        :rtype: torch.Tensor
        """
        return self.classifier(self.backbone(x))


import random as random

import numpy as np
import torch as torch
import torch.nn.functional as F
from torch import nn

from spuco.utils.random_seed import seed_randomness

# Acknowledgement to
# https://github.com/kuangliu/pytorch-cifar,
# https://github.com/BIGBALLON/CIFAR-ZOO,


class LeNet(nn.Module):
    """
    LeNet implementation.
    """
    def __init__(self, channel: int):
        """
        Initializes LeNet.

        :param channel: Number of input channels.
        :type channel: int
        """
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)
        super(LeNet, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(channel, 6, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(6, 16, kernel_size=5),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )
        self.fc_1 = nn.Linear(16 * 5 * 5, 120)
        self.fc_2 = nn.Linear(120, 84)
        self.representation_dim = 84

    def forward(self, x):       
        """
        Forward pass of the LeNet model.

        :param x: Input tensor.
        :type x: torch.Tensor

        :return: Output tensor.
        :rtype: torch.Tensor
        """
        x = self.features(x)
        x = x.view(x.size(0), -1)
        x = F.relu(self.fc_1(x))
        x = F.relu(self.fc_2(x))
        return x


from abc import ABC
from enum import Enum
from typing import Callable, Dict, List, Optional, Tuple
from tqdm import tqdm 
from spuco.datasets import BaseSpuCoCompatibleDataset

TRAIN_SPLIT = "train"
VAL_SPLIT= "val"
TEST_SPLIT = "test"

MASK_CORE = "core"
MASK_SPURIOUS = "spurious"

class SpuriousFeatureDifficulty(Enum):
    """
    Enumeration class for spurious feature difficulty levels.

    Each level corresponds to a combination of the magnitude and variance
    of the spurious feature.

    Magnitude definition of difficulty:
        Large Magnitude <-> Easy

        Medium Magnitude <-> Medium

        Small Magnitude <-> Hard

    Variance definition of difficulty:
        Low Variance <-> Easy

        Medium Variance <-> Medium

        High Variance <-> Hard
    """

    MAGNITUDE_LARGE = "magnitude_large"
    MAGNITUDE_MEDIUM = "magnitude_medium"
    MAGNITUDE_SMALL = "magnitude_small"
    VARIANCE_LOW = "variance_low"
    VARIANCE_MEDIUM = "variance_medium"
    VARIANCE_HIGH = "variance_high"

class SpuriousCorrelationStrength(Enum):
    UNIFORM = "unform"
    LINEAR = "linear"
    
class SourceData():
    """
    Class representing the source data.

    This class contains the input data and corresponding labels.
    """

    def __init__(self, data=None):
        """
        Initialize the SourceData object.

        :param data: The input data and labels.
        :type data: List[Tuple]
        """
        self.X = []
        self.labels = []
        self.spurious = []
        self.clean_labels = None
        self.core_feature_noise = None
        if data is not None:
            for x, label in tqdm(data):
                self.X.append(x)
                self.labels.append(label)

class BaseSpuCoDataset(BaseSpuCoCompatibleDataset, ABC):
    def __init__(
        self,
        root: str,
        num_classes: int,
        split: str = "train",
        transform: Optional[Callable] = None,
        verbose: bool = False,
    ):
        """
        Initializes the dataset.

        :param root: Root directory of the dataset.
        :type root: str

        :param num_classes: Number of classes in the dataset.
        :type num_classes: int

        :param split: Split of the dataset (e.g., "train", "test", "val"). Defaults to "train".
        :type split: str, optional

        :param transform: Optional transform to be applied to the data. Defaults to None.
        :type transform: Callable, optional

        :param verbose: Whether to print verbose information during dataset initialization. Defaults to False.
        :type verbose: bool, optional
        """
        
        super().__init__()
        self.root = root 
        self._num_classes = num_classes
        assert split == TRAIN_SPLIT or split == VAL_SPLIT or split == TEST_SPLIT, f"split must be one of {TRAIN_SPLIT}, {VAL_SPLIT}, {TEST_SPLIT}"
        self.split = split
        self.transform = transform
        self.verbose = verbose
        self.skip_group_validation = False

    def initialize(self):
        """
        Initializes the dataset.
        """
        # Load Data
        self.data, classes, spurious_classes = self.load_data()
        self.num_spurious = len(spurious_classes)
        
        # Group Partition
        self._group_partition = {}
        for i, group_label in enumerate(zip(self.data.labels, self.spurious)):
            if group_label not in self._group_partition:
                self._group_partition[group_label] = []
            self._group_partition[group_label].append(i)

        self._clean_group_partition = None
        if self.data.clean_labels is not None:
            self._clean_group_partition = {}
            for i, group_label in enumerate(zip(self.data.clean_labels, self.spurious)):
                if group_label not in self._clean_group_partition:
                    self._clean_group_partition[group_label] = []
                self._clean_group_partition[group_label].append(i)
            
        # Validate partition sizes
        if not self.skip_group_validation:
            for class_label in classes:
                for spurious_label in spurious_classes:
                    group_label = (class_label, spurious_label)
                    assert group_label in self._group_partition and len(self._group_partition[group_label]) > 0, f"No examples in {group_label}, considering reducing spurious correlation strength"

        # Group Weights
        self._group_weights = {}
        for key in self._group_partition.keys():
            self._group_weights[key] = len(self._group_partition[key]) / len(self.data.X)
                
    @property
    def group_partition(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Dictionary partitioning indices into groups

        :rtype: Dict[Tuple[int, int], List[int]]
        """
        return self._group_partition


    @property
    def clean_group_partition(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Dictionary partitioning indices into groups based on clean labels

        :rtype: Dict[Tuple[int, int], List[int]]
        """
        if self._clean_group_partition is None:
            return self._group_partition
        else:
            return self._clean_group_partition


    @property
    def group_weights(self) -> Dict[Tuple[int, int], float]:
        """
        Dictionary containing the fractional weights of each group

        :rtype: Dict[Tuple[int, int], float]
        """
        return self._group_weights


    @property
    def spurious(self) -> List[int]:
        """
        List containing spurious labels for each example

        :rtype: List[int]
        """
        return self.data.spurious


    @property
    def labels(self) -> List[int]:
        """
        List containing class labels for each example

        :rtype: List[int]
        """
        return self.data.labels


    @property
    def num_classes(self) -> int:
        """
        Number of classes

        :rtype: int
        """
        return self._num_classes

        
    def __getitem__(self, index):
        """
        Gets an item from the dataset.

        :param index: Index of the item to get.
        :type index: int
        :return: A tuple of (sample, target) where target is class_index of the target class.
        :rtype: tuple
        """
        if self.transform is None:
            return self.data.X[index], self.data.labels[index]
        else:
            return self.transform(self.data.X[index]), self.data.labels[index]
        
    def __len__(self):
        """
        Gets the length of the dataset.

        :return: Length of the dataset.
        :rtype: int
        """
        return len(self.data.X)



from typing import Dict, List, Tuple

from tqdm import tqdm
from wilds.datasets.wilds_dataset import WILDSDataset

from spuco.datasets import BaseSpuCoCompatibleDataset


class WILDSDatasetWrapper(BaseSpuCoCompatibleDataset):
    """
    Wrapper class that wraps WILDSDataset into a Dataset to be compatible with SpuCo.
    """
    def __init__(
        self,
        dataset: WILDSDataset,
        metadata_spurious_label: str,
        verbose=False
    ):
        """
        Wraps  WILDS Dataset into a Dataset object. 

        :param dataset: The source WILDS dataset
        :type dataset: WILDDataset
        :param metadata_spurious_label: String name of property in metadata_map corresponding to spurious target
        :type metadata_spurious_label: str 
        :param verbose: Show logs
        :type verbose: bool
        """
        super().__init__()

        self.dataset = dataset
        self._num_classes = dataset.n_classes 

        # Get index in meta data array corresponding to spurious target 
        spurious_target_idx = dataset.metadata_fields.index(metadata_spurious_label)

        # Get labels 
        self._labels = dataset.y_array.long().tolist()

        # Get spurious labels
        self._spurious = dataset.metadata_array[:, spurious_target_idx].long().tolist()

        # Create group partition using labels and spurious labels
        self._group_partition = {}
        for i, group_label in tqdm(
            enumerate(zip(self._labels, self._spurious)),
            desc="Partitioning data indices into groups",
            disable=not verbose,
            total=len(self.dataset)
        ):
            if group_label not in self._group_partition:
                self._group_partition[group_label] = []
            self._group_partition[group_label].append(i)
        
        # Set group weights based on group sizes
        self._group_weights = {}
        for group_label in self._group_partition.keys():
            self._group_weights[group_label] = len(self._group_partition[group_label]) / len(self.dataset)

    @property
    def group_partition(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Dictionary partitioning indices into groups
        """
        return self._group_partition 
    
    @property
    def group_weights(self) -> Dict[Tuple[int, int], float]:
        """
        Dictionary containing the fractional weights of each group
        """
        return self._group_weights
    
    @property
    def spurious(self) -> List[int]:
        """
        List containing spurious labels for each example
        """
        return self._spurious

    @property
    def labels(self) -> List[int]:
        """
        List containing class labels for each example
        """
        return self._labels
    
    @property
    def num_classes(self) -> int:
        """
        Number of classes
        """
        return self._num_classes

    def __getitem__(self, index):
        """
        Retrieves an item from the dataset.

        :param index: The index of the item.
        :type index: int
        :return: The item at the given index.
        """
        source_tuple = self.dataset.__getitem__(index)
        return (source_tuple[0], source_tuple[1])
    
    def __len__(self):
        """
        Returns the length of the dataset.

        :return: The length of the dataset.
        :rtype: int
        """
        return len(self.dataset)



from typing import Dict, List, Tuple

import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader


def convert_labels_to_partition(labels: List[int]) -> Dict[int, List[int]]:
    """
    Converts a list of labels into a partition dictionary.

    :param labels: List of labels.
    :type labels: List[int]
    :return: Partition dictionary mapping labels to their corresponding indices.
    :rtype: Dict[int, List[int]]
    """
    partition = {}
    for i, label in enumerate(labels):
        if label not in partition:
            partition[label] = []
        partition[label].append(i)
    return partition

def convert_partition_to_labels(partition: Dict[int, List[int]]) -> List[int]:
    """
    Converts a partition dictionary into a list of labels.

    :param partition: Partition dictionary mapping labels to their corresponding indices.
    :type partition: Dict[int, List[int]]
    :return: List of labels.
    :rtype: List[int]
    """
    labels = np.array([-1] * sum([len(partition[key]) for key in partition.keys()]))
    for key in partition.keys():
        labels[partition[key]] = key 
    return labels.tolist()

def label_examples(unlabled_dataloader: DataLoader, model: nn.Module, device: torch.device):
    """
    Labels examples using a trained model.

    :param unlabeled_dataloader: Dataloader containing unlabeled examples.
    :type unlabeled_dataloader: torch.utils.data.DataLoader
    :param model: Trained model for labeling examples.
    :type model: torch.nn.Module
    :param device: Device to use for computations.
    :type device: torch.device
    :return: List of predicted labels.
    :rtype: List[int]
    """
    labels = []
    for X in unlabled_dataloader:
        labels.append(torch.argmax(model(X.to(device)), dim=-1))
    labels = torch.cat(labels, dim=0)
    return labels.detach().cpu().tolist()

def pairwise_similarity(Z1: torch.tensor, Z2: torch.tensor, block_size: int = 1024):
    """
    Computes pairwise similarity between two sets of embeddings.

    :param Z1: Tensor containing the first set of embeddings.
    :type Z1: torch.tensor
    :param Z2: Tensor containing the second set of embeddings.
    :type Z2: torch.tensor
    :param block_size: Size of the blocks for computing similarity. Default is 1024.
    :type block_size: int
    :return: Pairwise similarity matrix.
    :rtype: np.array
    """
    similarity_matrices = []
    for i in range(Z1.shape[0] // block_size + 1):
        similarity_matrices_i = []
        e = Z1[i*block_size:(i+1)*block_size]
        for j in range(Z2.shape[0] // block_size + 1):
            e_t = Z2[j*block_size:(j+1)*block_size].t()
            similarity_matrices_i.append(
                np.array(
                torch.cosine_similarity(e[:, :, None], e_t[None, :, :]).detach().cpu()
                )
            )
        similarity_matrices.append(similarity_matrices_i)
    similarity_matrix = np.block(similarity_matrices)

    return similarity_matrix

def get_group_ratios(indices: List[int], group_partition: Dict[Tuple[int, int], List[int]]):
    """
    Returns the ratio of each group found in the given indices

    :param Z1: Tensor containing the first set of embeddings.
    :type Z1: torch.tensor
    :param Z2: Tensor containing the second set of embeddings.
    :type Z2: torch.tensor  
    """
    group_ratio = {}
    for key in group_partition.keys():
        group_ratio[key] = len([i for i in indices if i in group_partition[key]]) / len(group_partition[key])
    return group_ratio


from .base_group_inference import BaseGroupInference
from .cluster import Cluster, ClusterAlg
from .jtt_inference import JTTInference
from .eiil import EIIL
from .ssa import SSA
from .cnc_inference import CorrectNContrastInference
from .spare_inference import SpareInference


import random
from copy import deepcopy
from typing import Dict, List

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import Dataset

from spuco.evaluate import Evaluator
from spuco.robust_train import GroupBalanceBatchERM
from spuco.utils import convert_labels_to_partition
from spuco.utils.random_seed import seed_randomness


class SpareTrain(GroupBalanceBatchERM):
    """
    SpareTrain class for training a model using group balanced sampling
    """
    def __init__(
        self,
        model: nn.Module,
        trainset: Dataset,
        group_partition: Dict,
        sampling_powers: List,
        batch_size: int,
        optimizer: optim.Optimizer,
        num_epochs: int,
        device: torch.device = torch.device("cpu"),
        val_evaluator: Evaluator = None,
        verbose=False
    ):
        """
        Initializes GroupBalanceBatchERM.

        :param model: The PyTorch model to be trained.
        :type model: nn.Module
        :param trainset: The training dataset.
        :type trainset: Dataset
        :param group_partition: A dictionary mapping group labels to the indices of examples belonging to each group.
        :type group_partition: Dict
        :param batch_size: The batch size for training.
        :type batch_size: int
        :param optimizer: The optimizer used for training.
        :type optimizer: optim.Optimizer
        :param num_epochs: The number of training epochs.
        :type num_epochs: int
        :param device: The device to be used for training (default: CPU).
        :type device: torch.device
        :param verbose: Whether to print training progress (default: False).
        :type verbose: bool
        """
        
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)

        super().__init__(model=model, trainset=trainset, group_partition=group_partition, batch_size=batch_size, optimizer=optimizer, num_epochs=num_epochs, device=device, val_evaluator=val_evaluator, verbose=verbose)
        
        assert batch_size >= len(trainset.group_partition), "batch_size must be >= number of groups (Group DRO requires at least 1 example from each group)"
        
        self.class_partition = convert_labels_to_partition(trainset.labels)

        self.sampling_weights = []
        for key in self.group_partition.keys():
            self.sampling_weights.extend([1 / len(self.group_partition[key]) ** sampling_powers[key[0]]] * len(self.group_partition[key]))

        self.sampling_weights = np.array(self.sampling_weights)
        for key in self.class_partition.keys():
            # finds where class_partition[key] are in base indices
            indices = [i for i, x in enumerate(self.base_indices) if x in self.class_partition[key]]
            indices = np.array(indices)
            
            # normalize the sampling weights so that each class has the same total weight
            self.sampling_weights[indices] = self.sampling_weights[indices] / sum(self.sampling_weights[indices])
        
        self.sampling_weights = list(self.sampling_weights)

        
    def train_epoch(self, epoch: int):
        """
        Trains the model for a single epoch with a custom upsampled batch

        :param epoch: The current epoch number.
        :type epoch: int
        """
        self.trainer.sampler.indices = random.choices(
            population=self.base_indices,
            weights=self.sampling_weights, 
            k=len(self.trainer.trainset)
        )

        self.trainer.train_epoch(epoch)



import random
from typing import List

import numpy as np
import torch
from torch import nn, optim
from tqdm import tqdm

from spuco.group_inference import BaseGroupInference
from spuco.utils.random_seed import seed_randomness


class EIIL(BaseGroupInference):
    """
    Environment Inference for Invariant Learning: https://arxiv.org/abs/2010.07249
    """
    def __init__(
        self, 
        logits: torch.Tensor, 
        class_labels: List[int],
        num_steps: int, 
        lr: float,
        device: torch.device = torch.device("cpu"),
        verbose: bool = False

    ):
        """
        Initializes the EIILInference object.

        :param logits: The logits output of the model.
        :type logits: torch.Tensor
        :param class_labels: The class labels for each sample.
        :type class_labels: List[int]
        :param num_steps: The number of steps for training the soft environment assignment.
        :type num_steps: int
        :param lr: The learning rate for training.
        :type lr: float
        :param device: The device to use for training. Defaults to CPU.
        :type device: torch.device, optional
        :param verbose: Whether to print training progress. Defaults to False.
        :type verbose: bool, optional
        """
         
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)
        
        super().__init__()

        self.logits = logits 
        self.class_labels = class_labels
        self.num_steps = num_steps
        self.lr = lr
        self.device = device
        self.verbose = verbose

    def infer_groups(self):
        """
        Performs EIIL inference to infer group partitions.

        :return: The group partition based on EIIL inference.
        :rtype: Dict[Tuple[int, int], List[int]]
        """

        # Initialize
        scale = torch.tensor(1.).to(self.device).requires_grad_()
        train_criterion = nn.CrossEntropyLoss(reduction='none')
        loss = train_criterion(self.logits.to(self.device) * scale, torch.tensor(self.class_labels).long().to(self.device))
        env_w = torch.randn(len(self.logits)).to(self.device).requires_grad_()
        optimizer = optim.Adam([env_w], lr=self.lr)

        # Train assignment
        for i in tqdm(range(self.num_steps), disable=not self.verbose, desc="EIIL Inferring Groups"):
            # penalty for env a
            lossa = (loss.squeeze() * env_w.sigmoid()).mean()
            grada = torch.autograd.grad(lossa, [scale], create_graph=True)[0]
            penaltya = torch.sum(grada**2)
            # penalty for env b
            lossb = (loss.squeeze() * (1-env_w.sigmoid())).mean()
            gradb = torch.autograd.grad(lossb, [scale], create_graph=True)[0]
            penaltyb = torch.sum(gradb**2)
            # negate
            npenalty = - torch.stack([penaltya, penaltyb]).mean()
            optimizer.zero_grad()
            npenalty.backward(retain_graph=True)
            optimizer.step()

        # Sigmoid to get env assignment
        spurious_labels = env_w.sigmoid() > .5
        spurious_labels = spurious_labels.int().detach().cpu().numpy()
        
        # Partition using group labels to get group partition 
        group_partition = {}
        for i in range(len(spurious_labels)):
            group_label = (0, spurious_labels[i])
            if group_label not in group_partition:
                group_partition[group_label] = []
            group_partition[group_label].append(i)

        return group_partition



from .base_spuco_compatible_dataset import BaseSpuCoCompatibleDataset
from .base_spuco_dataset import SpuriousFeatureDifficulty, BaseSpuCoDataset, SourceData, TRAIN_SPLIT, VAL_SPLIT, TEST_SPLIT, SpuriousCorrelationStrength
from .base_spuco_dataset import MASK_CORE, MASK_SPURIOUS
from .spurious_target_dataset_wrapper import SpuriousTargetDatasetWrapper
from .group_labeled_dataset_wrapper import GroupLabeledDatasetWrapper
from .wilds_dataset_wrapper import WILDSDatasetWrapper
from .spuco_mnist import SpuCoMNIST
from .spuco_birds import SpuCoBirds
from .spuco_dogs import SpuCoDogs
from .spuco_animals import SpuCoAnimals


import random
from typing import Dict, List, Tuple

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset, SubsetRandomSampler
from tqdm import tqdm

from spuco.datasets import SpuriousTargetDatasetWrapper
from spuco.utils.random_seed import seed_randomness

from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from typing import Tuple, Optional


class Evaluator:
    def __init__(
        self,
        testset: Dataset, 
        group_partition: Dict[Tuple[int, int], List[int]],
        group_weights: Dict[Tuple[int, int], float],
        batch_size: int,
        model: nn.Module,
        sklearn_linear_model: Optional[Tuple[float, float, float, Optional[StandardScaler]]] = None,
        device: torch.device = torch.device("cpu"),
        verbose: bool = False
    ):
        """
        Initializes an instance of the Evaluator class.

        :param testset: Dataset object containing the test set.
        :type testset: Dataset

        :param group_partition: Dictionary object mapping group keys to a list of indices corresponding to the test samples in that group.
        :type group_partition: Dict[Tuple[int, int], List[int]]

        :param group_weights: Dictionary object mapping group keys to their respective weights.
        :type group_weights: Dict[Tuple[int, int], float]

        :param batch_size: Batch size for DataLoader.
        :type batch_size: int

        :param model: PyTorch model to evaluate.
        :type model: nn.Module

        :param sklearn_linear_model: Tuple representing the coefficients and intercept of the linear model from sklearn. Default is None.
        :type sklearn_linear_model: Optional[Tuple[float, float, float, Optional[StandardScaler]]], optional

        :param device: Device to use for computations. Default is torch.device("cpu").
        :type device: torch.device, optional

        :param verbose: Whether to print evaluation results. Default is False.
        :type verbose: bool, optional
        """
          
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        self.testloaders = {}
        self.group_partition = group_partition
        self.group_weights = group_weights
        self.model = model
        self.device = device
        self.verbose = verbose
        self.accuracies = None
        self.sklearn_linear_model = sklearn_linear_model
        self.n_classes = np.max(testset.labels) + 1

        # Create DataLoaders 

        # Group-Wise DataLoader
        for key in group_partition.keys():
            sampler = SubsetRandomSampler(group_partition[key])
            self.testloaders[key] = DataLoader(testset, batch_size=batch_size, sampler=sampler, num_workers=4, pin_memory=True, shuffle=False)
        
        # SpuriousTarget Dataloader
        spurious = torch.zeros(len(testset))
        for key in self.group_partition.keys():
            for i in self.group_partition[key]:
                spurious[i] = key[1]
        spurious_dataset = SpuriousTargetDatasetWrapper(dataset=testset, spurious_labels=spurious)
        self.spurious_dataloader = DataLoader(spurious_dataset, batch_size=batch_size, num_workers=4, pin_memory=True)

    def evaluate(self):
        """
        Evaluates the PyTorch model on the test dataset and computes the accuracy for each group.
        """
        self.model.eval()
        self.accuracies = {}
        for key in tqdm(sorted(self.group_partition.keys()), "Evaluating group-wise accuracy", ):
            if self.sklearn_linear_model:
                self.accuracies[key] = self._evaluate_accuracy_sklearn_logreg(self.testloaders[key])
            else:
                self.accuracies[key] = self._evaluate_accuracy(self.testloaders[key])
            if self.verbose:
                print(f"Group {key} Accuracy: {self.accuracies[key]}")
        return self.accuracies
    
    def _evaluate_accuracy(self, testloader: DataLoader):
        with torch.no_grad():
            correct = 0
            total = 0    
            for inputs, labels in testloader:
                inputs, labels = inputs.to(self.device), labels.to(self.device)
                outputs = self.model(inputs)
                predicted = torch.argmax(outputs, dim=1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()
            return 100 * correct / total
    
    def _evaluate_accuracy_sklearn_logreg(self, testloader: DataLoader):
        C, coef, intercept, scaler = self.sklearn_linear_model

        X_test, y_test = self._encode_testset(testloader)
        X_test = X_test.detach().cpu().numpy()
        y_test = y_test.detach().cpu().numpy()
        if scaler:
            X_test = scaler.transform(X_test)
        logreg = LogisticRegression(penalty='l1', C=C, solver="liblinear")
        # the fit is only needed to set up logreg
        logreg.fit(X_test[: self.n_classes], np.arange(self.n_classes))
        logreg.coef_ = coef
        logreg.intercept_ = intercept
        preds_test = logreg.predict(X_test)
        return (preds_test == y_test).mean()
    
    def _encode_testset(self, testloader):
        X_test = []
        y_test = []

        self.model.eval()
        with torch.no_grad():
            for inputs, labels in testloader:
                inputs, labels = inputs.to(self.device), labels.to(self.device)
                X_test.append(self.model.backbone(inputs))
                y_test.append(labels)
            return torch.cat(X_test), torch.cat(y_test)
        
    def evaluate_spurious_attribute_prediction(self):
        """
        Evaluates accuracy if the task was predicting the spurious attribute.
        """
        return self._evaluate_accuracy(self.spurious_dataloader)

    @property
    def worst_group_accuracy(self):
        """
        Returns the group with the lowest accuracy and its corresponding accuracy.

        :returns: A tuple containing the key of the worst-performing group and its corresponding accuracy.
        :rtype: tuple
        """
        if self.accuracies is None:
            print("Run evaluate() first")
            return None
        else:
            min_key = min(self.accuracies, key=self.accuracies.get)
            min_value = min(self.accuracies.values())
            return (min_key, min_value)
    
    @property
    def average_accuracy(self):
        """
        Returns the weighted average accuracy across all groups.

        :returns: The weighted average accuracy across all groups.
        :rtype: float
        """
        if self.accuracies is None:
            print("Run evaluate() first")
            return None
        else:
            accuracy = 0
            for key in self.group_partition.keys():
                accuracy += self.group_weights[key] * self.accuracies[key]
            return accuracy


import random
from typing import Any, Callable, Optional, Tuple

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import DataLoader, Dataset, Sampler
from tqdm import tqdm

from spuco.utils.random_seed import seed_randomness


class Trainer:
    def __init__(
            self,
            trainset: Dataset,
            model: nn.Module,
            batch_size: int,
            optimizer: optim.Optimizer,
            lr_scheduler: Optional[optim.lr_scheduler._LRScheduler] = None,
            max_grad_norm: Optional[float] = None,
            criterion: nn.Module = nn.CrossEntropyLoss(),
            forward_pass: Optional[Callable[[Any], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]] = None,
            sampler: Sampler = None,
            device: torch.device = torch.device("cpu"),
            verbose: bool = False
    ) -> None:
        """
        Initializes an instance of the Trainer class.

        :param trainset: The training set.
        :type trainset: torch.utils.data.Dataset
        :param model: The PyTorch model to train.
        :type model: torch.nn.Module
        :param batch_size: The batch size to use during training.
        :type batch_size: int
        :param optimizer: The optimizer to use for training.
        :type optimizer: torch.optim.Optimizer
        :param criterion: The loss function to use during training. Default is nn.CrossEntropyLoss().
        :type criterion: torch.nn.Module, optional
        :param forward_pass: The forward pass function to use during training. Default is None.
        :type forward_pass: Callable[[Any], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]], optional
        :param sampler: The sampler to use for creating batches. Default is None.
        :type sampler: torch.utils.data.Sampler, optional
        :param device: The device to use for computations. Default is torch.device("cpu").
        :type device: torch.device, optional
        :param verbose: Whether to print training progress. Default is False.
        :type verbose: bool, optional
        """
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        self.trainset = trainset
        self.model = model
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.max_grad_norm = max_grad_norm
        self.criterion = criterion
        self.batch_size = batch_size
        self.sampler = sampler
        self.verbose = verbose
        self.device = device
        
        if forward_pass is None:
            def forward_pass(self, batch):
                inputs, labels = batch
                inputs, labels = inputs.to(self.device), labels.to(self.device)
                outputs = self.model(inputs)
                loss = self.criterion(outputs, labels)
                return loss, outputs, labels
            self.forward_pass = forward_pass 
        else:
            self.forward_pass = forward_pass

        self.trainloader = DataLoader(
            self.trainset, 
            batch_size=self.batch_size, 
            shuffle=(self.sampler is None), 
            sampler=self.sampler,
            num_workers=4, 
            pin_memory=True
        )

    def train(self, num_epochs: int):
        """
        Trains for given number of epochs 

        :param num_epochs: Number of epochs to train for
        :type num_epochs: int
        """
        for epoch in range(num_epochs):
            self.train_epoch(epoch) 
            
    def train_epoch(self, epoch: int) -> None:
        """
        Trains the PyTorch model for 1 epoch

        :param epoch: epoch number that is being trained (only used by logging)
        :type epoch: int
        """
        self.model.train()
        with tqdm(self.trainloader, unit="batch", total=len(self.trainloader), disable=not self.verbose) as pbar:
            pbar.set_description(f"Epoch {epoch}")
            average_accuracy = 0.
            for batch in pbar:
                loss, outputs, labels = self.forward_pass(self, batch)
                accuracy = Trainer.compute_accuracy(outputs, labels)

                # backward pass and optimization
                self.optimizer.zero_grad()
                loss.backward()
                if self.max_grad_norm is not None:
                    nn.utils.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)
                if self.lr_scheduler is not None and isinstance(self.optimizer, optim.AdamW):
                    self.lr_scheduler.step()
                self.optimizer.step()

                pbar.set_postfix(loss=loss.item(), accuracy=f"{accuracy}%")
                average_accuracy += accuracy
                            
            if self.lr_scheduler is not None and not isinstance(self.optimizer, optim.AdamW):
                self.lr_scheduler.step()
                
            return average_accuracy / len(pbar)
    
    @staticmethod
    def compute_accuracy(outputs: torch.Tensor, labels: torch.Tensor) -> float:
        """
        Computes the accuracy of the PyTorch model.

        :param outputs: The predicted outputs of the model.
        :type outputs: torch.Tensor
        :param labels: The ground truth labels.
        :type labels: torch.Tensor
        :return: The accuracy of the model.
        :rtype: float
        """
        predicted = torch.argmax(outputs, dim=1)
        total = labels.size(0)
        correct = (predicted == labels).sum().item()
        return 100. * correct / total
    
    def get_trainset_outputs(self):
        """
        Gets output of model on trainset
        """
        with torch.no_grad():
            self.model.eval()
            eval_trainloader = DataLoader(
                dataset=self.trainset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=4, 
                pin_memory=True
            )
            with tqdm(eval_trainloader, unit="batch", total=len(self.trainloader), disable=not self.verbose) as pbar:
                outputs = []
                pbar.set_description("Getting Trainset Outputs")
                for input, _ in pbar:
                    outputs.append(self.model(input.to(self.device)))
                return torch.cat(outputs, dim=0)


from abc import ABC, abstractmethod
from typing import Dict, List, Tuple
import pickle 

class BaseGroupInference(ABC):
    """
    BaseGroupInference abstract class for inferring group partitions.
    """
    def __init__(self):
        """
        Initializes BaseGroupInference.
        """
        pass 

    @abstractmethod
    def infer_groups(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Abstract method for inferring group partitions.

        :return: Dictionary mapping group tuples to indices of examples belonging to each group.
        """
        pass 

    def process_cluster_partition(self, cluster_partition: Dict, class_index: int):
        """
        Processes cluster partition:
        - Converts keys from clusters into (class, spurious) format
        - Converts class indices from class-wise clustering into global (actual trainset) indices

        :param cluster_partition: Dictionary mapping cluster labels to indices of examples.
        :param class_index: Index of the class being processed.
        :return: Processed group partition mapping group tuples to indices of examples belonging to each group.
        """
        assert self.class_partition is not None, "self.class_partition must be defined for processing"
        group_partition = {}
        for new_cluster_label, cluster_label in enumerate(sorted(cluster_partition.keys())):
            group_partition[(class_index, new_cluster_label)] = [self.class_partition[class_index][i] for i in cluster_partition[cluster_label]]
        return group_partition
    
    @staticmethod
    def save_group_partition(group_partition: Dict[Tuple[int, int], List[int]], prefix: str):
        with open(f"{prefix}_group_partition.pkl", "wb") as f:
            pickle.dump(group_partition, f)



from .submodular import FacilityLocation, lazy_greedy
from .exemplar_cluster import closest_exemplar, cluster_by_exemplars
from .misc import *
from .custom_indices_sampler import CustomIndicesSampler
from .trainer import Trainer 
from .random_seed import set_seed
from .bert_transform import initialize_bert_transform


from typing import List, Dict, Tuple

import torch
from torch import nn, optim
from torch.utils.data import Dataset

from spuco.evaluate import Evaluator
from spuco.robust_train import BaseRobustTrain
from spuco.utils import CustomIndicesSampler, Trainer
from spuco.utils.random_seed import seed_randomness
import random 
import numpy as np 

class UpSampleERM(BaseRobustTrain):
    """
    UpSampleERM class for training a model by upsampling all groups to size of largest group. 
    """
    def __init__(
        self,
        model: nn.Module,
        trainset: Dataset,
        batch_size: int,
        optimizer: optim.Optimizer,
        num_epochs: int,
        group_partition: Dict[Tuple[int, int], List[int]],
        criterion=nn.CrossEntropyLoss(), 
        device: torch.device = torch.device("cpu"),
        val_evaluator: Evaluator = None,
        verbose=False
    ):  
        """
        Initializes UpSampleERM.

        :param model: The PyTorch model to be trained.
        :type model: nn.Module
        :param trainset: The training dataset.
        :type trainset: Dataset
        :param batch_size: The batch size for training.
        :type batch_size: int
        :param optimizer: The optimizer used for training.
        :type optimizer: optim.Optimizer
        :param num_epochs: The number of training epochs.
        :type num_epochs: int
        :param indices: A list of indices specifying the samples to be shown to the model in 1 epoch.
        :type indices: List[int]
        :param criterion: The loss criterion used for training (default: CrossEntropyLoss).
        :type criterion: nn.Module
        :param device: The device to be used for training (default: CPU).
        :type device: torch.device
        :param verbose: Whether to print training progress (default: False).
        :type verbose: bool
        """

        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)
        
        super().__init__(val_evaluator=val_evaluator, verbose=verbose)

        self.num_epochs = num_epochs

        len_max_group = max([len(group_partition[key]) for key in group_partition.keys()])
        self.indices = []
        for key in group_partition.keys():
            group_reordered_indices = [group_partition[key][i] for i in torch.randperm(len(group_partition[key])).tolist()]
            group_indices = []
            while len(group_indices) < len_max_group:
                group_indices.extend(group_reordered_indices)
            self.indices.extend(group_indices[:len_max_group])

        self.trainer = Trainer(
            trainset=trainset,
            model=model,
            batch_size=batch_size,
            optimizer=optimizer,
            criterion=criterion,
            sampler=CustomIndicesSampler(indices=self.indices, shuffle=True),
            verbose=verbose,
            device=device
        )


import random
from enum import Enum
from typing import Dict, List, Optional

import numpy as np
import torch
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from tqdm import tqdm

from spuco.group_inference import BaseGroupInference
from spuco.utils import (cluster_by_exemplars, convert_labels_to_partition,
                         convert_partition_to_labels, pairwise_similarity)
from spuco.utils.random_seed import seed_randomness, get_seed


class ClusterAlg(Enum):
    KMEANS = "kmeans"
    KMEDOIDS = "kmedoids"

class Cluster(BaseGroupInference):
    """
    Clustering-based Group Inference
    """
    def __init__(
        self,
        Z: torch.Tensor,
        class_labels: Optional[List[int]] = None,
        cluster_alg: ClusterAlg = ClusterAlg.KMEANS,
        num_clusters: int = -1,
        max_clusters: int = -1,
        device: torch.device = torch.device("cpu"), 
        verbose: bool = False
    ):
        """
        Initializes the Cluster object.

        :param Z: The input tensor for clustering.
        :type Z: torch.Tensor
        :param class_labels: Optional list of class labels for class-wise clustering. Defaults to None.
        :type class_labels: Optional[List[int]], optional
        :param cluster_alg: The clustering algorithm to use. Defaults to ClusterAlg.KMEANS.
        :type cluster_alg: ClusterAlg, optional
        :param num_clusters: The number of clusters to create. Defaults to -1.
        :type num_clusters: int, optional
        :param max_clusters: The maximum number of clusters to consider. Defaults to -1.
        :type max_clusters: int, optional
        :param device: The device to run the clustering on. Defaults to torch.device("cpu").
        :type device: torch.device, optional
        :param verbose: Whether to display progress and logging information. Defaults to False.
        :type verbose: bool, optional
        """
         
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        super().__init__()

        # Argument Validation 
        if num_clusters < 2:
            if max_clusters < 2:
                raise ValueError("At least one of num_clusters and max_clusters must be valid i.e. >= 2")
        elif max_clusters >= 2:
            raise ValueError("num_clusters and max_clusters are mutually exclusive")

        # Partition into classes if given labels to compute clusters class-wise
        self.class_partition = {}
        if class_labels is None:
            self.class_partition[-1] = range(len(Z))
        else:
            self.class_partition = convert_labels_to_partition(class_labels)
        
        # Initialize variables
        self.cluster_alg = cluster_alg
        self.num_clusters = num_clusters 
        self.max_clusters = max_clusters
        self.device = device
        self.verbose = verbose

        # Processing Z
        self.Z = Z
        if cluster_alg == ClusterAlg.KMEANS:
            self.Z = self.Z.detach().cpu().numpy()

    def infer_groups(self) -> Dict[int, List[int]]:
        """
        Infers the group partition based on the clustering results.

        :return: The group partition where each key is a cluster label and the value is a list of indices belonging to that cluster.
        :rtype: Dict[int, List[int]]
        """ 
        # Get class-wise group partitions
        cluster_partitions = [] 
        for class_label in tqdm(self.class_partition.keys(), disable=not self.verbose, desc="Clustering class-wise"):
            Z = self.Z[self.class_partition[class_label]]
            if self.num_clusters < 0:
                partition, scores = self.silhouette(Z)
            elif self.cluster_alg == ClusterAlg.KMEANS:
                _, partition = self.kmeans(Z, num_clusters=self.num_clusters)
            else:
                similarity_matrix = pairwise_similarity(Z.to(self.device), Z.to(self.device))
                _, partition = self.kmedoids(Z, similiarity_matrix=similarity_matrix)
            cluster_partitions.append(partition)

        # Merge class-wise group partitions into one dictionary
        group_partition = {}
        for class_index, partition in zip(self.class_partition.keys(), cluster_partitions):
            group_partition.update(self.process_cluster_partition(partition, class_index))
        return group_partition
    
    def silhouette(self, Z):
        """
        Uses the silhouette score to determine the optimal number of clusters and perform clustering based on self.cluster_alg.

        :param Z: The input data for clustering.
        :type Z: torch.Tensor

        :return: The cluster partition based on the optimal number of clusters.
        :rtype: List[int]
        """

        silhouette_scores = []
        partitions = []

        similarity_matrix = None 
        if self.cluster_alg == ClusterAlg.KMEDOIDS:
            similarity_matrix = pairwise_similarity(Z.to(self.device), Z.to(self.device))
        
        # Iterate through possible num_clusters
        for num_clusters in range(2, self.max_clusters+1):
            # Cluster using num_clusters
            cluster_labels, cluster_partition = None, None 
            if self.cluster_alg == ClusterAlg.KMEANS:
                cluster_labels, cluster_partition = self.kmeans(Z, num_clusters=num_clusters)
                silhouette_scores.append(silhouette_score(Z, cluster_labels)) 
            else: 
                cluster_labels, cluster_partition = self.kmedoids(Z, similiarity_matrix=similarity_matrix, num_clusters=num_clusters)
                silhouette_scores.append(silhouette_score(Z.cpu().numpy(), cluster_labels)) 
            partitions.append(cluster_partition)
             
            if self.verbose:
                print("For n_clusters =", num_clusters,
                    "The average silhouette_score is :", silhouette_scores[-1])
        # Pick best num_clusters
        best_partition_idx = np.argmax(silhouette_scores)
        return partitions[best_partition_idx], silhouette_scores[best_partition_idx]
    
    def kmeans(self, Z, num_clusters: int =-1):
        """
        Performs K-means clustering on the input data.

        :param Z: The input data for clustering.
        :type Z: torch.Tensor
        :param num_clusters: The number of clusters to create. If not specified, the value from the object will be used.
        :type num_clusters: int, optional

        :return: The cluster labels and partition based on the K-means clustering.
        :rtype: Tuple[np.ndarray, List[List[int]]]
        """
        # if num_clusters not passed, use value from object
        if num_clusters < 0:
            num_clusters = self.num_clusters

        # K-Means 
        clusterer = KMeans(n_clusters=num_clusters,
                            random_state=get_seed(),
                            n_init=10)
        cluster_labels = clusterer.fit_predict(Z)
        return cluster_labels, convert_labels_to_partition(cluster_labels.astype(int).tolist())

    def kmedoids(self, Z, similiarity_matrix: torch.Tensor, num_clusters=-1):
        """
        Performs K-medoids clustering on the input data.

        :param Z: The input data for clustering.
        :type Z: torch.Tensor
        :param similiarity_matrix: The similarity matrix for pairwise similarities between data points.
        :type similiarity_matrix: torch.Tensor
        :param num_clusters: The number of clusters to create. If not specified, the value from the object will be used.
        :type num_clusters: int, optional

        :return: The cluster labels and partition based on the K-medoids clustering.
        :rtype: Tuple[np.ndarray, List[List[int]]]
        """
        # if num_clusters not passed, use value from object
        if num_clusters < 0:
            num_clusters = self.num_clusters

        # K-Medoids greedy (cluster by exemplars)
        cluster_partition = cluster_by_exemplars(
            similarity_matrix=similiarity_matrix, 
            num_exemplars=num_clusters, 
            verbose=self.verbose
        )

        return convert_partition_to_labels(cluster_partition), cluster_partition


import random 

class Singleton(object):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Singleton, cls).__new__(cls)
        return cls.instance

def set_seed(new_seed: int):
    """
    Set the seed for the SpuCo module.

    :param new_seed: The new seed value to set.
    :type new_seed: int
    """
    seed_class = Singleton()
    seed_class.seed = new_seed 

def get_seed():
    """
    Get the seed value of the SpuCo module.

    :return: The seed value.
    :rtype: int
    """
    seed_class = Singleton()
    if not hasattr(seed_class, "seed"):
        return random.randint(0, 10000000)
    else:
        return seed_class.seed
    
def seed_randomness(random_module=None, torch_module=None, numpy_module=None):
    """
    Seed the randomness of the specified modules.

    :param random_module: The random module. Default is None.
    :type random_module: Optional[ModuleType]
    :param torch_module: The torch module. Default is None.
    :type torch_module: Optional[ModuleType]
    :param numpy_module: The numpy module. Default is None.
    :type numpy_module: Optional[ModuleType]
    """
    seed = get_seed()
    if torch_module is not None:
        torch_module.backends.cudnn.deterministic = True 
        torch_module.backends.cudnn.benchmark = False 
        torch_module.manual_seed(seed)
    if numpy_module is not None:
        numpy_module.random.seed(seed)
    if random_module is not None:
        random_module.seed(seed)


import random
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import Dataset

from spuco.group_inference import BaseGroupInference
from spuco.utils import Trainer
from spuco.utils.random_seed import seed_randomness


class CorrectNContrastInference(BaseGroupInference):
    """
    Correct-n-Contrast Inference: https://proceedings.mlr.press/v162/zhang22z.html
    """
    def __init__(
        self,
        trainset: Dataset,
        model: nn.Module,
        batch_size: int,
        optimizer: optim.Optimizer,
        num_epochs: int, 
        device: torch.device = torch.device("cpu"),
        verbose: bool = False
    ):
        """
        Initializes the CorrectNContrastInference object.

        :param trainset: The training dataset.
        :type trainset: Dataset
        :param model: The model for training.
        :type model: nn.Module
        :param batch_size: The batch size for training.
        :type batch_size: int
        :param optimizer: The optimizer for training.
        :type optimizer: optim.Optimizer
        :param num_epochs: The number of epochs for training.
        :type num_epochs: int
        :param device: The device to use for training. Defaults to CPU.
        :type device: torch.device, optional
        :param verbose: Whether to print training progress. Defaults to False.
        :type verbose: bool, optional
        """
         
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)
        
        super().__init__()
        
        self.trainer = Trainer(
            trainset=trainset,
            model=model, 
            batch_size=batch_size,
            optimizer=optimizer,
            device=device,
            verbose=verbose
        )
        self.num_epochs = num_epochs

    def infer_groups(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Performs Correct-n-Contrast inference to infer group partitions.

        :return: The group partition based on Correct-n-Contrast inference.
        :rtype: Dict[Tuple[int, int], List[int]]
        """
        
        self.trainer.train(self.num_epochs)
        
        spurious = torch.argmax(self.trainer.get_trainset_outputs(), dim=-1).cpu().tolist()

        # CNC only wants to group based on spurious attribute, but for consistency in API
        # we need a tuple as key, hence group_label set to (0, spurious) 
        group_partition = {}
        for i in range(len(self.trainer.trainset)):
            group_label = (0, spurious[i])
            if group_label not in group_partition:
                group_partition[group_label] = []
            group_partition[group_label].append(i)
            
        return group_partition




import os
import pickle
import random
from typing import Callable, Optional

import numpy as np
import requests
import torch
import torchvision.transforms as transforms
from PIL import Image
from tqdm import tqdm

from spuco.datasets import (TRAIN_SPLIT, MASK_CORE, MASK_SPURIOUS, 
                            BaseSpuCoDataset, SourceData,
                            SpuCoBirds, SpuCoDogs)
from spuco.utils.random_seed import seed_randomness

MASK_URL = "https://ucla.box.com/s/v3tdeoc6zhvhk4hytlxsh2nmql54o740"
MASK_FNAME = "spuco_animals_masks.pkl"

class SpuCoAnimals(BaseSpuCoDataset):
    """
    Next, we introduce SpuCoAnimals, a large-scale vision dataset curated from ImageNet with two realistic spurious correlations. 

    SpuCoAnimals has 4 classes: 

    - landbirds
    - waterbirds
    - small dog breeds
    - big dog breeds.

    Waterbirds and Landbirds are spuriously correlated with *water* and *land* backgrounds, respectively. Small dogs and big dogs are spuriously correlated with *indoor* and *outdoor* backgrounds, respectively.
    """

    def __init__(
        self,
        root: str,
        download: bool = True,
        label_noise: float = 0.0,
        split: str = TRAIN_SPLIT,
        transform: Optional[Callable] = None,
        mask_type: Optional[str] = None,
        verbose: bool = False
    ):
        """
        Initializes the dataset.

        :param root: Root directory of the dataset.
        :type root: str

        :param download: Whether to download the dataset. Defaults to True.
        :type download: bool, optional

        :param label_noise: The amount of label noise to apply. Defaults to 0.0.
        :type label_noise: float, optional

        :param split: The split of the dataset. Defaults to TRAIN_SPLIT.
        :type split: str, optional

        :param transform: Optional transform to be applied to the data. Defaults to None.
        :type transform: Callable, optional

        :param verbose: Whether to print verbose information during dataset initialization. Defaults to False.
        :type verbose: bool, optional
        """

        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        super().__init__(
            root=root, 
            split=split,
            transform=transform,
            num_classes=4,
            verbose=verbose
        )
        self.download = download
        self.label_noise = label_noise
        self.base_transform = transforms.Compose([
            transforms.Resize((256, 256)),
            transforms.CenterCrop((224, 224)),
            transforms.ToTensor()
        ])
        
        self.mask_type = mask_type
        
        # Don't have examples in "all" groups, skip validation
        self.skip_group_validation = True 
        
    def load_data(self) -> SourceData:
        """
        Loads SpuCoAnimals and sets spurious labels, label noise.

        :return: The spurious correlation dataset.
        :rtype: SourceData, List[int], List[int]
        """
        if self.verbose:
            print("Loading SpuCoBirds")
        self.data = SpuCoBirds(
            root=self.root,
            download=self.download,
            label_noise=self.label_noise,
            split=self.split,
            transform=self.transform,
            verbose=self.verbose   
        ).load_data()[0]
        
        if self.verbose:
            print("Loading SpuCoDogs")
        self.dogs_data = SpuCoDogs(
            root=self.root,
            download=self.download,
            label_noise=self.label_noise,
            split=self.split,
            transform=self.transform,
            verbose=self.verbose   
        ).load_data()[0]
        
        # If masks required, download and load
        if self.mask_type is not None:
            self.mask_fname = os.path.join(self.root, MASK_FNAME)
            
            # Download masks (if needed)
            if not os.path.exists(self.mask_fname):
                if self.verbose:
                    print(f"{self.mask_fname} not found, downloading instead.")
                self._download_masks()
                
            if self.verbose: 
                print("Loading Masks")
                
            with open(self.mask_fname, "rb") as f:
                self.masks = pickle.load(f)
        
        self.data.core_feature_noise = None
        self.data.X.extend(self.dogs_data.X)
        self.data.labels.extend([label + 2 for label in self.dogs_data.labels])
        self.data.spurious.extend([label + 2 for label in self.dogs_data.spurious])
        if self.data.clean_labels is not None:   
            self.data.clean_labels.extend([label + 2 for label in self.dogs_data.clean_labels])   
        
        return self.data, list(range(4)), list(range(4))
    
    def __getitem__(self, index):
        """
        Gets an item from the dataset.

        :param index: Index of the item to get.
        :type index: int
        :return: A tuple of (sample, target) where target is class_index of the target class.
        :rtype: tuple
        """
        
        image = self.base_transform(self.load_image(self.data.X[index]))
        label = self.data.labels[index]
        if self.transform is None:
            return image, label
        else:
            return self.transform(image), label
        
    def load_image(self, filename: str):
        image = Image.open(filename).convert("RGB")
        
        if self.mask_type is None:
            return image 
        
        image = np.array(image)
        mask_index = int(filename.split("/")[-1].split(".")[0])
        spurious_mask = np.moveaxis(np.stack([self.masks[mask_index]] * 3), 0, 2)
        if self.mask_type == MASK_SPURIOUS:
            image = image * spurious_mask
        elif self.mask_type == MASK_CORE:
            image = image * np.invert(spurious_mask)
        else:
            raise ValueError(f"Mask Type must be one of None, {MASK_CORE}, {MASK_SPURIOUS}")
        
        return Image.fromarray(image)    
    
    def _download_masks(self):
        response = requests.get(MASK_URL, stream=True)
        response.raise_for_status()

        with open(self.mask_fname, "wb") as file:
            for chunk in tqdm(response.iter_content(chunk_size=1024), total=136800, desc="Downloading Masks for SpuCoAnimals", unit="KB"):
                file.write(chunk)


from abc import abstractmethod
from typing import Dict, List, Tuple

from torch.utils.data import Dataset


class BaseSpuCoCompatibleDataset(Dataset):
    """
    Base class for all SpuCo Compatible Datasets
    """
    def __init__(self):
        super().__init__()

    @property
    @abstractmethod
    def group_partition(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Dictionary partitioning indices into groups
        """
        pass 
    
    @property
    @abstractmethod
    def group_weights(self) -> Dict[Tuple[int, int], float]:
        """
        Dictionary containing the fractional weights of each group
        """
        pass 
    
    @property
    @abstractmethod
    def spurious(self) -> List[int]:
        """
        List containing spurious labels for each example
        """
        pass 

    @property
    @abstractmethod
    def labels(self) -> List[int]:
        """
        List containing class labels for each example
        """
        pass
    
    @property
    @abstractmethod
    def num_classes(self) -> int:
        """
        Number of classes
        """
        pass


import itertools
import random
from copy import deepcopy
from enum import Enum
from typing import Callable, List, Optional

import matplotlib.cm as cm
import numpy as np
import torch
import torchvision
import torchvision.transforms as T
from torch.utils.data import Subset
from tqdm import tqdm

from spuco.datasets import (TEST_SPLIT, TRAIN_SPLIT, VAL_SPLIT,
                            BaseSpuCoDataset, SourceData,
                            SpuriousFeatureDifficulty)
from spuco.datasets.spuco_mnist_config import config
from spuco.utils.random_seed import seed_randomness


class ColourMap(Enum):
    HSV = "hsv"
    
class SpuCoMNIST(BaseSpuCoDataset):
    """
    A dataset consisting of images from the MNIST dataset
    with added spurious features to create a spurious MNIST dataset.
    """

    def __init__(
        self,
        root: str,
        spurious_feature_difficulty: SpuriousFeatureDifficulty,
        classes: List[List[int]],
        spurious_correlation_strength = 0.0,
        label_noise: float = 0.0,
        core_feature_noise: float = 0.0,
        color_map: ColourMap = ColourMap.HSV,
        split: str = TRAIN_SPLIT,
        transform: Optional[Callable] = None,
        verbose: bool = False,
        download: bool = True,
    ):
        """
        Initializes the SpuCoMNIST dataset.

        :param root: The root directory of the dataset.
        :type root: str

        :param spurious_feature_difficulty: The difficulty level of the spurious feature.
        :type spurious_feature_difficulty: SpuriousFeatureDifficulty

        :param classes: The list of class labels for each digit.
        :type classes: List[List[int]]

        :param spurious_correlation_strength: The strength of the spurious feature correlation. Default is 0.
        :type spurious_correlation_strength: float, optional

        :param label_noise: The amount of label noise to apply. Default is 0.0.
        :type label_noise: float, optional

        :param core_feature_noise: The amount of noise to add to the core features. Default is 0.0.
        :type core_feature_noise: float, optional

        :param color_map: The color map to use. Default is ColourMap.HSV.
        :type color_map: ColourMap, optional

        :param split: The dataset split to load. Default is "train".
        :type split: str, optional

        :param transform: The data transformation function. Default is None.
        :type transform: Optional[Callable], optional

        :param verbose: Whether to print verbose information during dataset initialization. Default is False.
        :type verbose: bool, optional

        :param download: Whether to download the dataset. Default is True.
        :type download: bool, optional
        """
        
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        super().__init__(
            root=root, 
            split=split,
            transform=transform,
            num_classes=len(classes),
            verbose=verbose,
        )
        
        self.label_noise = label_noise
        assert self.label_noise >= 0.0 and self.label_noise <= 1, "Invalid label_noise value"
        self.core_feature_noise = core_feature_noise
        assert self.core_feature_noise >= 0.0 and self.core_feature_noise <= 1, "Invalid core_feature_noise value"
        self.spurious_correlation_strength = spurious_correlation_strength
        self.spurious_feature_difficulty = spurious_feature_difficulty
        self.classes = classes
        self.colors = self.init_colors(color_map)
        self.download = download

    def load_data(self) -> SourceData:
        """
        Loads the MNIST dataset and generates the spurious correlation dataset.

        :return: The spurious correlation dataset.
        :rtype: SourceData
        """
        self.mnist = torchvision.datasets.MNIST(
            root=self.root, 
            train=self.split != TEST_SPLIT, 
            download=self.download,
            transform=T.Compose([
                T.ToTensor(),
                T.Lambda(lambda x: torch.cat([x, x, x], dim=0))  # convert grayscale to RGB
            ])
        )
            
        # Return predetermined train / val split
        if self.split == TRAIN_SPLIT:
            self.mnist = Subset(
                dataset=self.mnist, 
                indices=[i for i in range(len(self.mnist)) if i not in config[VAL_SPLIT]]
            )
        elif self.split == VAL_SPLIT:
            self.mnist = Subset(
                dataset=self.mnist, 
                indices=config[VAL_SPLIT]
            )

        self.data = SourceData(self.mnist)
        
        # Validate Classes
        assert SpuCoMNIST.validate_classes(self.classes), "Classes should be disjoint and only contain elements 0<= label <= 9"

        # Get New Labels
        kept = []
        for i, label in enumerate(self.data.labels):
            for class_idx, latent_class in enumerate(self.classes):
                if label in latent_class:
                    self.data.labels[i] = class_idx
                    kept.append(i)
        
        self.data.X = [self.data.X[i] for i in kept]
        self.data.labels = [self.data.labels[i] for i in kept]
        
        # Partition indices by new labels
        self.partition = {}
        for i, label in enumerate(self.data.labels):
            if label not in self.partition:
                self.partition[label] = []
            self.partition[label].append(i)
        
        # Train / Val: Add spurious correlation iteratively for each class
        self.data.spurious = [-1] * len(self.data.X)
        if self.split == TRAIN_SPLIT or (self.split == VAL_SPLIT and self.spurious_correlation_strength != 0):
            assert self.spurious_correlation_strength != 0, f"spurious correlation strength must be specified and > 0 for split={TRAIN_SPLIT}"

            # Determine label noise idx
            self.is_noisy_label = torch.zeros(len(self.data.X))
            if self.label_noise > 0:
                self.data.clean_labels = deepcopy(self.data.labels)
                self.is_noisy_label[torch.randperm(len(self.data.X))[:int(self.label_noise * len(self.data.X))]] = 1

            # Determine feature noise idx
            self.data.core_feature_noise = torch.zeros(len(self.data.X))
            if self.core_feature_noise > 0:
                self.data.core_feature_noise[torch.randperm(len(self.data.X))[:int(self.core_feature_noise * len(self.data.X))]] = 1

            for label in tqdm(self.partition.keys(), desc="Adding spurious feature", disable=not self.verbose):

                # Get spurious correlation strength for this class
                spurious_correlation_strength = self.spurious_correlation_strength
                if type(self.spurious_correlation_strength) == list:
                    spurious_correlation_strength = self.spurious_correlation_strength[label]

                # Randomly permute and choose which points will have spurious feature (avoids issue of sampling leading to 
                # too many examples having spurious ---> no examples for some groups)
                is_spurious = torch.zeros(len(self.partition[label]))
                is_spurious[torch.randperm(len(self.partition[label]))[:int(spurious_correlation_strength * len(self.partition[label]))]] = 1

                # Get what the other labels could be for this class (all but spurious)
                other_labels = [x for x in range(len(self.classes)) if x != label]

                # Determine background of every example 
                background_label = torch.tensor([label if is_spurious[i] else other_labels[i % len(other_labels)] for i in range(len(self.partition[label]))])
                background_label = background_label[torch.randperm(len(background_label))]

                # Create and apply background for all examples
                for i, idx in enumerate(self.partition[label]):
                    self.data.spurious[idx] = background_label[i].item()
                    background = SpuCoMNIST.create_background(self.spurious_feature_difficulty, self.colors[self.data.spurious[idx]])

                    # Feature noise is a random mask applied to core feature
                    core_feature_noise = torch.ones_like(self.data.X[idx]) >= 1.0 # default noise is no noise
                    if self.data.core_feature_noise[idx]:
                        core_feature_noise = (torch.randn_like(self.data.X[idx][0, :, :]) > 0.25).unsqueeze(dim=0).repeat(3, 1, 1)
                    self.data.X[idx] = self.data.X[idx] * core_feature_noise
                    self.data.X[idx] = (background * (self.data.X[idx] == 0)) + self.data.X[idx]
                    
                    # If noisy label
                    if self.is_noisy_label[idx]:
                        self.data.labels[idx] = random.choice(other_labels)


        # Test / Val: Create spurious balanced test set
        else:
            for label in tqdm(self.partition.keys(), desc="Adding spurious feature", disable=not self.verbose):
                # Generate balanced background labels
                background_label = torch.tensor([i % len(self.classes) for i in range(len(self.partition[label]))])
                background_label = background_label[torch.randperm(len(background_label))]

                # Create and apply background for all examples
                for i, idx in enumerate(self.partition[label]):
                    self.data.spurious[idx] = background_label[i].item()
                    background = SpuCoMNIST.create_background(self.spurious_feature_difficulty, self.colors[self.data.spurious[idx]])
                    self.data.X[idx] = (background * (self.data.X[idx] == 0)) + self.data.X[idx]   

        # Return data, list containing all class labels, list containing all spurious labels
        return self.data, range(len(self.classes)), range(len(self.classes))

    def init_colors(self, color_map: ColourMap) -> List[List[float]]:
        """
        Initializes the color values for the spurious features.

        :param color_map: The color map to use for the spurious features. Should be a value from the `ColourMap`
            enum class.
        :type color_map: ColourMap
        
        :return: The color values for the spurious features.
        :rtype: List[List[float]]
        """
        color_map = cm.get_cmap(color_map.value)
        cmap_vals = np.arange(0, 1, step=1 / len(self.classes))
        colors = []
        for i in range(len(self.classes)):
            rgb = color_map(cmap_vals[i])[:3]
            rgb = [float(x) for x in np.array(rgb)]
            colors.append(rgb)
        # Append black as no-spurious background
        colors.append([0., 0., 0.])
        return colors
    
    @staticmethod
    def validate_classes(classes: List[List[int]]) -> bool:
        """
        Validates that the classes provided to the `SpuCoMNIST` dataset are disjoint and only contain integers
        between 0 and 9.

        :param classes: The classes to be included in the dataset, where each element is a list of integers
            representing the digits to be included in a single class.
        :type classes: List[List[int]]

        :return: Whether the classes are valid.
        :rtype: bool
        """
        sets = [set(latent_class) for latent_class in classes]

        for i in range(len(sets)):
            if any([x < 0 or x > 9 for x in sets[i]]):
                return False
            for j in range(i + 1, len(sets)):
                if sets[i].intersection(sets[j]):
                    return False
        return True

    @staticmethod
    def create_background(spurious_feature_difficulty: SpuriousFeatureDifficulty, hex_code: str) -> torch.Tensor:
        """
        Generates a tensor representing a background image with a specified spurious feature difficulty and hex code color.

        :param spurious_feature_difficulty: The difficulty level of the spurious feature to add to the background image.
        :type spurious_feature_difficulty: SpuriousFeatureDifficulty

        :param hex_code: The hex code of the color to use for the background image.
        :type hex_code: str

        :return: A tensor representing the generated background image.
        :rtype: torch.Tensor
        """
        background = SpuCoMNIST.rgb_to_mnist_background(hex_code)
        if spurious_feature_difficulty == SpuriousFeatureDifficulty.MAGNITUDE_LARGE:
            return background
        elif spurious_feature_difficulty == SpuriousFeatureDifficulty.MAGNITUDE_MEDIUM:
            unmask_points = torch.tensor(list(itertools.product(range(4), range(4))))
            mask = SpuCoMNIST.compute_mask(unmask_points)
        elif spurious_feature_difficulty == SpuriousFeatureDifficulty.MAGNITUDE_SMALL:
            unmask_points = torch.tensor(list(itertools.product(range(2), range(2))))
            mask = SpuCoMNIST.compute_mask(unmask_points)
        elif spurious_feature_difficulty == SpuriousFeatureDifficulty.VARIANCE_LOW:
            unmask_points = torch.tensor(list(itertools.product(range(7), range(7))))
            mask = SpuCoMNIST.compute_mask(unmask_points)
        elif spurious_feature_difficulty == SpuriousFeatureDifficulty.VARIANCE_MEDIUM:
            all_points = torch.tensor(list(itertools.product(range(14), range(14))))
            unmask_points = all_points[torch.randperm(len(all_points))[:49]]
            mask = SpuCoMNIST.compute_mask(unmask_points)
        elif spurious_feature_difficulty == SpuriousFeatureDifficulty.VARIANCE_HIGH:
            all_points = torch.tensor(list(itertools.product(range(28), range(28))))
            unmask_points = all_points[torch.randperm(len(all_points))[:49]]
            mask = SpuCoMNIST.compute_mask(unmask_points)
        return background * mask
    
    @staticmethod
    def compute_mask(unmask_points: torch.Tensor) -> torch.Tensor:
        """
        Computes a binary mask based on the unmasked points.

        :param unmask_points: The coordinates of the unmasked points.
        :type unmask_points: torch.Tensor
        :return: The binary mask with 1s at the unmasked points and 0s elsewhere.
        :rtype: torch.Tensor
        """
        rows = torch.tensor([point[0] for point in unmask_points])
        cols = torch.tensor([point[1] for point in unmask_points])
        mask = torch.zeros((3,28,28))
        mask[:, rows, cols] = 1.
        return mask
    
    @staticmethod
    def rgb_to_mnist_background(rgb: List[float]) -> torch.Tensor:
        """
        Converts an RGB color to a MNIST background tensor.

        :param rgb: The RGB color values.
        :type rgb: List[float]
        :return: The MNIST background tensor with the specified RGB color.
        :rtype: torch.Tensor
        """
        return torch.tensor(rgb).unsqueeze(1).unsqueeze(2).repeat(1, 28, 28)  


import random
from typing import Iterator, List

import numpy as np
import torch
from torch.utils.data import Sampler

from spuco.utils.random_seed import seed_randomness


class CustomIndicesSampler(Sampler[int]):
    """
    Samples from the specified indices (pass indices - upsampled, downsampled, group balanced etc. to this class)
    Default is no shuffle.
    """
    def __init__(
        self, 
        indices: List[int],
        shuffle: bool = False,
    ):
        """
        Samples elements from the specified indices.

        :param indices: The list of indices to sample from.
        :type indices: list[int]
        :param shuffle: Whether to shuffle the indices. Default is False.
        :type shuffle: bool, optional
        """
         
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)

        self.indices = indices
        self.shuffle = shuffle

    def __iter__(self) -> Iterator[int]:
        """
        Returns an iterator over the sampled indices.

        :return: An iterator over the sampled indices.
        :rtype: iterator[int]
        """
        if self.shuffle:
            random.shuffle(self.indices)
        return iter(self.indices)
    
    def __len__(self) -> int:
        """
        Returns the number of sampled indices.

        :return: The number of sampled indices.
        :rtype: int
        """
        return len(self.indices)


import random

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
from .dfr import DFR
from spuco.datasets import BaseSpuCoCompatibleDataset, GroupLabeledDatasetWrapper
from spuco.models import SpuCoModel
from spuco.utils.random_seed import seed_randomness
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from typing import List, Optional, Dict
import random

class DISPEL(DFR):
    def __init__(
        self,
        group_labeled_set:  GroupLabeledDatasetWrapper,
        model: SpuCoModel,
        labeled_valset_size: float = 0.5,
        C_range: List[float] = [1., 0.7, 0.3, 0.1, 0.07, 0.03, 0.01],
        s_range: List[float] = [1.0, 0.9, 0.8, 0.7],
        class_weight_options: Optional[List[Dict]] = None,
        validation_set: Optional[GroupLabeledDatasetWrapper] = None,
        alpha_range : List[float] = [1.0],
        size_of_mixed: Optional[int] = None,
        n_lin_models: int = 20, 
        data_for_scaler: Optional[Dataset] = None,
        group_unlabeled_set: Optional[Dataset] = None,
        device: torch.device = torch.device("cpu"),
        verbose: bool = False,
    ):
        """
        Initializes the DISPEL object.

        :param group_labeled_set: The group-labeled dataset.
        :type group_labeled_set: Dataset
        :param model: The base model.
        :type model: SpucoModel
        :param n_lin_models: Number of linear models to average.
        :type trainset: int
        :param labeled_valset_size: The ratio of the labeled data to be used for validation if no validation set is given. 
        :type labeled_valset_size: float
        :param C_range: Options of C, which is the inverse of l1 regularization strength as in sklearn.
        :type C_range: list
        :param s_range: Options of s, which is the weights assigned to the group-labeled data when generating mixed data.
        :type s_range: list
        :param alpha_range: Options of alpha, which is the probabilty of mixing data 
        :type alpha_range: list
        :param size_of_mixed: size of the mixed dataset. 
        :type size of_mixed: int
        :param group_unlabeled_set: group unlabeled dataset. If provided, it will be included in the group unbalanced dataset.
        :type group_unlabeled_set: Dataset
        :param class_weight_options: options for class weight.
        :type class_weight_options: list
        :param validation_set: data used for hyperparameter selection. If not provided, half of the group labeled data will be used.
        :type validation_set: GroupLabeledDatasetWrapper
        :param data_for_scaler: Data used for fitting the sklearn scaler. If not provided, group labeled data will be used.
        :type data_for_scaler: Dataset
        """

        super().__init__(
        group_labeled_set=group_labeled_set,
        model=model,
        labeled_valset_size=labeled_valset_size,
        C_range=C_range,
        class_weight_options=class_weight_options,
        validation_set=validation_set,
        n_lin_models=n_lin_models, 
        device=device,
        verbose=verbose,
        data_for_scaler=data_for_scaler)

        self.alpha_range = alpha_range
        self.s_range = s_range
        self.group_unlabeled_set = group_unlabeled_set
        self.size_of_mixed = size_of_mixed

    def train_single_model(self, alpha, s, C, X_labeled, y_labeled, g_labeled, class_weight):
        """
        Trains a single model.

        :param alpha: Trade-off parameter between accuracy and fairness.
        :type alpha: float

        :param s: Sensitivity parameter for fairness regularization.
        :type s: float

        :param C: Regularization parameter C for the SVM model.
        :type C: float

        :param X_labeled: Labeled training features.
        :type X_labeled: numpy.ndarray

        :param y_labeled: Labeled training labels.
        :type y_labeled: numpy.ndarray

        :param g_labeled: Labeled training group labels.
        :type g_labeled: numpy.ndarray

        :param class_weight: Weight associated with each class.
        :type class_weight: dict or 'balanced', optional
        """
        # sample the maximal group balanced set from the group labeled data
        group_partition = []
        for g in range(np.max(g_labeled)+1):
            group_partition.append(np.where(g_labeled==g)[0])
        min_size = np.min([len(g) for g in group_partition])
        X_balanced = []
        y_balanced = []
        for g in group_partition:
            indices = np.random.choice(g, size=min_size, replace=False)
            X_balanced.append(X_labeled[indices])
            y_balanced.append(y_labeled[indices])
        X_balanced = np.concatenate(X_balanced)
        y_balanced = np.concatenate(y_balanced)

        # construct the group unbalanced set, which is the union of all group labeled and unlabeled data
        if self.X_unlabeled is not None:
            X_unbalanced = np.concatenate((X_labeled, self.X_unlabeled))
            y_unbalanced = np.concatenate((y_labeled, self.y_unlabeled))
        else:
            X_unbalanced = X_labeled
            y_unbalanced = y_labeled
        
        X_class_balanced = []
        y_class_balanced = []
        n_classes = np.max(y_unbalanced)+1
        if self.size_of_mixed is None:
            c_min = np.min([len(np.where(y_unbalanced == c)[0]) for c in range(n_classes) ])
            for c in range(n_classes):
                ids = np.where(y_unbalanced == c)[0]
                ids = np.random.choice(ids, size=c_min, replace=False)
                X_class_balanced.append(X_unbalanced[ids])  
                y_class_balanced.append(y_unbalanced[ids])  
        else:  
            for c in range(n_classes):
                ids = np.where(y_unbalanced == c)[0]
                ids = np.random.choice(ids, size=self.size_of_mixed//n_classes, replace=True)
                X_class_balanced.append(X_unbalanced[ids])
                y_class_balanced.append(y_unbalanced[ids])  
        X_unbalanced = np.vstack(X_class_balanced)
        y_unbalanced = np.hstack(y_class_balanced)

        X_mixed = []
        y_mixed = []
        for i in range(X_unbalanced.shape[0]):
            if np.random.rand() > alpha:
                X_mixed.append(X_unbalanced[i])
                y_mixed.append(y_unbalanced[i])
            else:
                # sample another example in the same class from the balanced dataset
                y = y_unbalanced[i]
                idx = np.where(y_balanced == y)[0]
                j = np.random.choice(idx)
                X_mixed.append(X_balanced[j] * s + X_unbalanced[i] * (1 - s))
                y_mixed.append(y)
        X_mixed = np.array(X_mixed)
        y_mixed = np.array(y_mixed)

        logreg = LogisticRegression(penalty='l1', C=C, solver="liblinear", class_weight=class_weight)
        logreg.fit(X_mixed, y_mixed)
        
        return logreg.coef_, logreg.intercept_
    

    def train_multiple_model(self, alpha, s, C, X_labeled_train, y_labeled_train, g_labeled_train, class_weight):
        """
        Trains the DFR model.

        :param alpha: Trade-off parameter between accuracy and fairness.
        :type alpha: float

        :param s: Sensitivity parameter for fairness regularization.
        :type s: float

        :param C: Regularization parameter C for the SVM model.
        :type C: float

        :param X_labeled_train: Labeled training features.
        :type X_labeled_train: numpy.ndarray

        :param y_labeled_train: Labeled training labels.
        :type y_labeled_train: numpy.ndarray

        :param g_labeled_train: Labeled training group labels.
        :type g_labeled_train: numpy.ndarray

        :param class_weight: Weight associated with each class.
        :type class_weight: dict or 'balanced', optional
        """
        coefs, intercepts = [], []
        for i in range(self.n_lin_models):
            coef_, intercept_ = self.train_single_model(alpha, s, C, X_labeled_train, y_labeled_train, g_labeled_train, class_weight)
            coefs.append(coef_)
            intercepts.append(intercept_)
        return np.mean(coefs, axis=0), np.mean(intercepts, axis=0)
    

    def hyperparam_selection(self, X_labeled_train, y_labeled_train, g_labeled_train, X_labeled_val, y_labeled_val, g_labeled_val):
        """
        Performs hyperparameter selection for the DFR model.

        :param X_labeled_train: Labeled training features.
        :type X_labeled_train: numpy.ndarray

        :param y_labeled_train: Labeled training labels.
        :type y_labeled_train: numpy.ndarray

        :param g_labeled_train: Labeled training group labels.
        :type g_labeled_train: numpy.ndarray

        :param X_labeled_val: Labeled validation features.
        :type X_labeled_val: numpy.ndarray

        :param y_labeled_val: Labeled validation labels.
        :type y_labeled_val: numpy.ndarray

        :param g_labeled_val: Labeled validation group labels.
        :type g_labeled_val: numpy.ndarray
        """
        best_wg_acc = -1
        if self.verbose:
            print('Searching for best hyperparameters ...')
        for alpha in self.alpha_range:
            for s in self.s_range:
                for C in self.C_range:
                    for class_weight in self.class_weight_options:
                        coef, intercept = self.train_single_model(alpha, s, C, X_labeled_train, y_labeled_train, g_labeled_train, class_weight)
                        # coef, intercept = self.train_multiple_model(C, X_labeled_train, y_labeled_train, g_labeled_train)
                        wg_acc = self.evaluate_worstgroup_acc(C, coef, intercept, X_labeled_val, y_labeled_val, g_labeled_val)
                        if wg_acc > best_wg_acc:
                            best_wg_acc = wg_acc
                            self.best_alpha = alpha
                            self.best_s = s
                            self.best_C = C
                            self.best_class_weight = class_weight
                        if self.verbose:
                            print('alpha {} | s {} |C {} | class weight {} : Val Worst-Group Accuracy: {}'.format(alpha, s, C, class_weight, wg_acc))
                            print('Best alpha {} | s {} |C {} | class weight {}: Best Val Worst-Group Accuracy: {}'.format(self.best_alpha, self.best_s, self.best_C, self.best_class_weight, best_wg_acc))
            
    
    def train(self):
        """
        Last Layer Retraining.
        """
        
        if self.verbose:
            print('Encoding data ...')

        X_labeled, y_labeled, g_labeled = self.encode_dataset(self.group_labeled_set)
        X_labeled = X_labeled.detach().cpu().numpy()
        y_labeled = y_labeled.detach().cpu().numpy()
        g_labeled = g_labeled.detach().cpu().numpy()

        # Load the unlabeled dataset if it is given
        if self.group_unlabeled_set:
            self.X_unlabeled, self.y_unlabeled = self.encode_dataset(self.group_unlabeled_set)
            self.X_unlabeled = self.X_unlabeled.detach().cpu().numpy()
            self.y_unlabeled = self.y_unlabeled.detach().cpu().numpy()
        else:
            self.X_unlabeled, self.y_unlabeled = None, None

        # Standardize features
        if self.preprocess:
            self.scaler = StandardScaler()
            if self.data_for_scaler:
                X_scaler, _ = self.encode_dataset(self.data_for_scaler)
                X_scaler = X_scaler.detach().cpu().numpy()
                self.scaler.fit(X_scaler)
            else:
                self.scaler = StandardScaler()
                self.scaler.fit(X_labeled)
            X_labeled = self.scaler.transform(X_labeled)
            if self.group_unlabeled_set:
                self.X_unlabeled = self.scaler.transform(self.X_unlabeled)

        # If validation set is not provided, split labeled data into training and validation data
        # Otherwise, use the given validation set 
        if self.validation_set is None:
            n_labeled = X_labeled.shape[0]
            ids = {i for i in range(n_labeled)}
            ids_val = set(random.sample(ids, int(self.labeled_valset_size*n_labeled)))
            ids_train = ids - ids_val
            ids_val = np.array(list(ids_val))
            ids_train = np.array(list(ids_train))

            X_labeled_train, y_labeled_train, g_labeled_train = X_labeled[ids_train], y_labeled[ids_train], g_labeled[ids_train]
            X_labeled_val, y_labeled_val, g_labeled_val = X_labeled[ids_val], y_labeled[ids_val], g_labeled[ids_val]
        else:
            X_labeled_train, y_labeled_train, g_labeled_train = X_labeled, y_labeled, g_labeled
            X_labeled_val, y_labeled_val, g_labeled_val = self.encode_dataset(self.validation_set)
            X_labeled_val = X_labeled_val.detach().cpu().numpy()
            y_labeled_val = y_labeled_val.detach().cpu().numpy()
            g_labeled_val = g_labeled_val.detach().cpu().numpy()
            if self.preprocess:
                X_labeled_val = self.scaler.transform(X_labeled_val)    

        if self.class_weight_options is None:
            n_class = np.max(y_labeled_val) + 1
            self.class_weight_options = [{c: 1 for c in range(n_class)}]

        self.hyperparam_selection(X_labeled_train, y_labeled_train, g_labeled_train, X_labeled_val, y_labeled_val, g_labeled_val)
        coef, intercept = self.train_multiple_model(self.best_alpha, self.best_s, self.best_C, X_labeled_train, y_labeled_train, g_labeled_train, self.best_class_weight)
        self.linear_model = (self.best_C, coef, intercept, self.scaler)


import random
from enum import Enum
from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
from torchvision.models import resnet50, ResNet50_Weights, resnet18, ResNet18_Weights

from spuco.models import MLP, Bert, DistilBert, LeNet, SpuCoModel
from spuco.utils.random_seed import seed_randomness


class Identity(nn.Module):
    def __init__(self):
        super().__init__()
    def forward(self, x):
        return x

class SupportedModels(Enum):
    """
    Enum listing all supported models.
    """
    MLP = "mlp"
    LeNet = "lenet"
    BERT = "bert"
    DistilBERT = "distilbert"
    ResNet18 = "resnet18"
    ResNet50 = "resnet50"

def model_factory(arch: str, input_shape: Tuple[int, int, int], num_classes: int, pretrained: bool = True):
    """
    Factory function to create a SpuCoModel based on the specified architecture.

    :param arch: The architecture name.
    :type arch: str
    :param input_shape: The shape of the input data in the format (channels, height, width).
    :type input_shape: Tuple[int, int, int]
    :param num_classes: The number of output classes.
    :type num_classes: int
    :param pretrained: Whether to load pretrained weights. Default is True.
    :type pretrained: bool
    :return: A SpuCoModel instance.
    :rtype: SpuCoModel
    :raises NotImplementedError: If the specified architecture is not supported.
    """
    
    seed_randomness(random_module=random, torch_module=torch, numpy_module=np)
    arch = SupportedModels(arch)
    channel = input_shape[0]
    image_size = (input_shape[1], input_shape[2])
    backbone = None
    representation_dim = -1 
    
    if arch == SupportedModels.MLP:
        backbone = MLP(channel * image_size[0] * image_size[1])
        representation_dim = backbone.representation_dim
    elif arch == SupportedModels.LeNet: 
        backbone = LeNet(channel=channel)
        representation_dim = backbone.representation_dim
    elif arch == SupportedModels.BERT:
        backbone = Bert.from_pretrained('bert-base-uncased')
        representation_dim = backbone.d_out
    elif arch == SupportedModels.DistilBERT:
        backbone = DistilBert.from_pretrained('distilbert-base-uncased')
        representation_dim = backbone.d_out
    elif arch == SupportedModels.ResNet18:
        if pretrained:
            backbone = resnet18(weights=ResNet18_Weights.IMAGENET1K_V1)
        else:
            backbone = resnet18(weights=None)
        representation_dim = backbone.fc.in_features
        backbone.fc = Identity()
    elif arch == SupportedModels.ResNet50:
        if pretrained:
            backbone = resnet50(weights=ResNet50_Weights.IMAGENET1K_V1)
        else:
            backbone = resnet50(weights=None)
        representation_dim = backbone.fc.in_features
        backbone.fc = Identity()
    else:
        raise NotImplemented(f"Model {arch} not supported currently")
    return SpuCoModel(
        backbone=backbone, 
        representation_dim=representation_dim, 
        num_classes=num_classes
    )




from typing import Dict, List

import numpy as np
from tqdm import tqdm

from spuco.utils.submodular import FacilityLocation, lazy_greedy

def cluster_by_exemplars(similarity_matrix, num_exemplars, verbose=False) -> Dict[int, List[int]]:
    """
    Returns a dictionary mapping exemplar index to a list of indices.

    :param similarity_matrix: The similarity matrix.
    :type similarity_matrix: numpy.ndarray
    :param num_exemplars: The number of exemplars to select.
    :type num_exemplars: int
    :param verbose: Whether to print progress information.
    :type verbose: bool, optional
    :return: A dictionary mapping exemplar index to a list of indices.
    :rtype: dict[int, list[int]]]
    """
    submodular_function = FacilityLocation(D=similarity_matrix, V=range(len(similarity_matrix)))
    exemplar_indices, _ = lazy_greedy(F=submodular_function, V=range(len(similarity_matrix)), B=num_exemplars, verbose=verbose)
    clusters = {}

    for exemplar_index in exemplar_indices:
        clusters[exemplar_index] = []

    for index in tqdm(range(len(similarity_matrix)), desc="Sorting samples by exemplar", disable=not verbose):
        exemplar_index, _  = closest_exemplar(index, exemplar_indices, similarity_matrix)
        clusters[exemplar_index].append(index)

    return clusters

def closest_exemplar(sample_index, exemplar_indices, similarity_matrix):
    """
    Finds the closest exemplar to a given sample index.

    :param sample_index: The index of the sample.
    :type sample_index: int
    :param exemplar_indices: The indices of the exemplars.
    :type exemplar_indices: list[int]
    :param similarity_matrix: The similarity matrix.
    :type similarity_matrix: numpy.ndarray
    :return: The index of the closest exemplar and the similarity score.
    :rtype: tuple[int, float]
    """

    max_similarity = np.NINF
    best_exemplar_index = -1

    for curr_exemplar_index in exemplar_indices:
        if similarity_matrix[sample_index][curr_exemplar_index] > max_similarity:
            max_similarity = similarity_matrix[sample_index][curr_exemplar_index]
            best_exemplar_index = curr_exemplar_index 
    
    return best_exemplar_index, max_similarity



import random
from typing import Any, Callable, Optional, Tuple

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import DataLoader, Dataset, Sampler
from tqdm import tqdm

from spuco.datasets import GroupLabeledDatasetWrapper
from spuco.evaluate import Evaluator
from spuco.robust_train import BaseRobustTrain
from spuco.models import SpuCoModel
from spuco.utils.random_seed import seed_randomness


class CorrectNContrastTrain(BaseRobustTrain):
    """
    CorrectNContrastTrain class for training a model using CNC's 
    Cross Entropy + modified Supervised Contrastive Learning loss.
    """
    def __init__(
        self,
        trainset: GroupLabeledDatasetWrapper,
        model: SpuCoModel,
        batch_size: int,
        optimizer_encoder: optim.Optimizer,
        optimizer_classifier: optim.Optimizer,
        num_pos: int, 
        num_neg: int,
        num_epochs: int,
        lambda_ce: float,
        temp: float,
        device: torch.device = torch.device("cpu"),
        accum: int = 32, 
        val_evaluator: Evaluator = None,
        verbose: bool = False  
    ):
        """
        Initializes CorrectNContrastTrain.

        :param trainset: The training dataset containing group-labeled samples.
        :type trainset: GroupLabeledDatasetWrapper
        :param model: The SpuCoModel to be trained.
        :type model: SpuCoModel
        :param batch_size: The batch size for training.
        :type batch_size: int
        :param optimizer: The optimizer used for training.
        :type optimizer: optim.Optimizer
        :param num_pos: The number of positive examples for contrastive loss.
        :type num_pos: int
        :param num_neg: The number of negative examples for contrastive loss.
        :type num_neg: int
        :param num_epochs: The number of training epochs.
        :type num_epochs: int
        :param lambda_ce: The weight of the regular cross-entropy loss.
        :type lambda_ce: float
        :param temp: The temperature the regular cross-entropy loss.
        :type temp: float
        :param device: The device to be used for training (default: CPU).
        :type device: torch.device
        :param verbose: Whether to print training progress (default: False).
        :type verbose: bool
        """
        
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)
        
        super().__init__(val_evaluator=val_evaluator, verbose=verbose)
    
        self.num_epochs = num_epochs 

        def forward_pass(self, batch):
            """
            Custom forward pass function for Correct & Contrast training.

            :param batch: A batch of input samples.
            :return: The loss, model outputs, and labels.
            """
            # Unpack inputs and move to correct device
            inputs, labels, groups = batch 
            inputs, labels, groups = inputs.to(self.device), labels.to(self.device), groups.to(self.device)

            # Regular Cross Entropy Loss
            outputs = self.model(inputs)
            loss = lambda_ce * self.criterion(outputs, labels)

            # Contrastive Loss
            for anchor_idx in range(len(batch)):
                # Only computed with majority examples as anchor
                if labels[anchor_idx] != groups[anchor_idx]:
                    continue 
                pos_idx = []
                neg_idx = []
                for i in range(len(inputs)):
                    if len(pos_idx) < num_pos: # Positives = same label, but different group
                        if labels[i] == labels[anchor_idx] and groups[i] != groups[anchor_idx]:
                            pos_idx.append(i)

                    if len(neg_idx) < num_neg: # Negatives = different label, but same spurious attribute
                        if labels[i] != labels[anchor_idx] and groups[i] == groups[anchor_idx]:
                            neg_idx.append(i)
                if len(pos_idx) > 0 and len(neg_idx) > 0:
                    anchor = self.model.backbone(torch.unsqueeze(inputs[anchor_idx], dim=0))
                    pos = self.model.backbone(inputs[pos_idx])
                    neg = self.model.backbone(inputs[neg_idx])
                    pos_sim = torch.exp(torch.cosine_similarity(anchor, pos)/temp)
                    pos_sum_sim = torch.sum(pos_sim) 
                    neg_sum_sim = torch.sum(torch.exp(torch.cosine_similarity(anchor, neg)/temp))
                    sup_cl_loss = torch.sum(torch.log(pos_sim / (pos_sum_sim + neg_sum_sim)))
                    loss += (1-lambda_ce) * sup_cl_loss / len(pos_idx)

            return loss, outputs, labels

        self.trainer = CNCTrainer(
            trainset=trainset,
            model=model,
            batch_size=batch_size,
            optimizer_1=optimizer_encoder,
            optimizer_2=optimizer_classifier,
            accum_1=1,
            accum_2=accum,
            forward_pass=forward_pass,
            verbose=verbose,
            device=device
        )

class CNCTrainer:
    def __init__(
            self,
            trainset: Dataset,
            model: nn.Module,
            batch_size: int,
            optimizer_1: optim.Optimizer,
            optimizer_2: optim.Optimizer,
            accum_1: int,
            accum_2: int,
            lr_scheduler: Optional[optim.lr_scheduler._LRScheduler] = None,
            max_grad_norm: Optional[float] = None,
            criterion: nn.Module = nn.CrossEntropyLoss(),
            forward_pass: Optional[Callable[[Any], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]] = None,
            sampler: Sampler = None,
            device: torch.device = torch.device("cpu"),
            verbose: bool = False
    ) -> None:
        """
        Initializes an instance of the Trainer class.

        :param trainset: The training set.
        :type trainset: torch.utils.data.Dataset
        :param model: The PyTorch model to train.
        :type model: torch.nn.Module
        :param batch_size: The batch size to use during training.
        :type batch_size: int
        :param optimizer: The optimizer to use for training.
        :type optimizer: torch.optim.Optimizer
        :param criterion: The loss function to use during training. Default is nn.CrossEntropyLoss().
        :type criterion: torch.nn.Module, optional
        :param forward_pass: The forward pass function to use during training. Default is None.
        :type forward_pass: Callable[[Any], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]], optional
        :param sampler: The sampler to use for creating batches. Default is None.
        :type sampler: torch.utils.data.Sampler, optional
        :param device: The device to use for computations. Default is torch.device("cpu").
        :type device: torch.device, optional
        :param verbose: Whether to print training progress. Default is False.
        :type verbose: bool, optional
        """
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        self.trainset = trainset
        self.model = model
        self.batch_size = batch_size
        self.optimizer_1 = optimizer_1
        self.optimizer_2 = optimizer_2
        self.accum_1 = accum_1
        self.accum_2 = accum_2
        self.lr_scheduler = lr_scheduler
        self.max_grad_norm = max_grad_norm
        self.criterion = criterion
        self.batch_size = batch_size
        self.sampler = sampler
        self.verbose = verbose
        self.device = device
        
        if forward_pass is None:
            def forward_pass(self, batch):
                inputs, labels = batch
                inputs, labels = inputs.to(self.device), labels.to(self.device)
                outputs = self.model(inputs)
                loss = self.criterion(outputs, labels)
                return loss, outputs, labels
            self.forward_pass = forward_pass 
        else:
            self.forward_pass = forward_pass

        self.trainloader = DataLoader(
            self.trainset, 
            batch_size=self.batch_size, 
            shuffle=(self.sampler is None), 
            sampler=self.sampler,
            num_workers=4, 
            pin_memory=True
        )

    def train(self, num_epochs: int):
        """
        Trains for given number of epochs 

        :param num_epochs: Number of epochs to train for
        :type num_epochs: int
        """
        for epoch in range(num_epochs):
            self.train_epoch(epoch) 
            
    def train_epoch(self, epoch: int) -> None:
        """
        Trains the model for 1 epoch using CNC method

        :param epoch: epoch number that is being trained (only used by logging)
        :type epoch: int
        """
        self.model.train()
        self.optimizer_1.zero_grad()
        self.optimizer_2.zero_grad()
        with tqdm(self.trainloader, unit="batch", total=len(self.trainloader), disable=not self.verbose) as pbar:
            pbar.set_description(f"Epoch {epoch}")
            average_accuracy = 0.
            idx_batch = 1
            for batch in pbar:
                loss, outputs, labels = self.forward_pass(self, batch)
                accuracy = CNCTrainer.compute_accuracy(outputs, labels)

                # backward pass and optimization
                
                loss.backward()
                if self.max_grad_norm is not None:
                    nn.utils.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)
                if self.lr_scheduler is not None and isinstance(self.optimizer_1, optim.AdamW):
                    self.lr_scheduler.step()
                if idx_batch % self.accum_1 == 0 or idx_batch == (len(self.trainloader)):
                    self.optimizer_1.step()
                    self.optimizer_1.zero_grad()
                if idx_batch % self.accum_2 == 0 or idx_batch == (len(self.trainloader)):
                    self.optimizer_2.step()
                    self.optimizer_2.zero_grad()

                # TODO: check if step should be called every batch or every epoch
                if self.lr_scheduler is not None and not isinstance(self.optimizer_1, optim.AdamW):
                    self.lr_scheduler.step()

                pbar.set_postfix(loss=loss.item(), accuracy=f"{accuracy}%")
                average_accuracy += accuracy
                idx_batch += 1
            return average_accuracy / len(pbar)
    
    @staticmethod
    def compute_accuracy(outputs: torch.Tensor, labels: torch.Tensor) -> float:
        """
        Computes the accuracy of the PyTorch model.

        :param outputs: The predicted outputs of the model.
        :type outputs: torch.Tensor
        :param labels: The ground truth labels.
        :type labels: torch.Tensor
        :return: The accuracy of the model.
        :rtype: float
        """
        predicted = torch.argmax(outputs, dim=1)
        total = labels.size(0)
        correct = (predicted == labels).sum().item()
        return 100. * correct / total
    
    def get_trainset_outputs(self):
        """
        Gets output of model on trainset
        """
        with torch.no_grad():
            self.model.eval()
            eval_trainloader = DataLoader(
                dataset=self.trainset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=4, 
                pin_memory=True
            )
            with tqdm(eval_trainloader, unit="batch", total=len(self.trainloader), disable=not self.verbose) as pbar:
                outputs = []
                pbar.set_description("Getting Trainset Outputs")
                for input, _ in pbar:
                    outputs.append(self.model(input.to(self.device)))
                return torch.cat(outputs, dim=0)


import random
from typing import Callable

import numpy as np
import torch
from torch import nn, optim

from spuco.datasets import GroupLabeledDatasetWrapper
from spuco.evaluate import Evaluator 
from spuco.robust_train import BaseRobustTrain
from spuco.utils import CustomIndicesSampler, Trainer
from spuco.utils.random_seed import seed_randomness


class GroupWeightedLoss(nn.Module):
    """
    A module for computing group-weighted loss.
    """
    def __init__(
        self, 
        criterion: Callable[[torch.tensor, torch.tensor], torch.tensor],
        num_groups: int,
        group_weight_lr: float = 0.01,
        device: torch.device = torch.device("cpu"),
    ):
        """
        Initializes GroupWeightedLoss.

        :param criterion: The loss criterion function.
        :type criterion: Callable[[torch.tensor, torch.tensor], torch.tensor]
        :param num_groups: The number of groups to consider.
        :type num_groups: int
        :param group_weight_lr: The learning rate for updating group weights (default: 0.01).
        :type group_weight_lr: float
        :param device: The device on which to perform computations. Defaults to CPU.
        :type device: torch.device
        """
        super(GroupWeightedLoss, self).__init__()
        self.criterion = criterion
        self.device = device
        self.num_groups = num_groups
        self.group_weights = torch.ones(self.num_groups).to(self.device)
        self.group_weights.data = self.group_weights.data / self.group_weights.data.sum()
        self.group_weight_lr = group_weight_lr

    def forward(self, outputs, labels, groups):
        """
        Computes the group-weighted loss.
        """
        # compute loss for different groups and update group weights
        loss = self.criterion(outputs, labels)
        group_loss = torch.zeros(self.num_groups).to(self.device)
        for i in range(self.num_groups):
            if (groups==i).sum() > 0:
                group_loss[i] += loss[groups==i].mean()
        self.update_group_weights(group_loss)

        # compute weighted loss
        loss = group_loss * self.group_weights
        loss = loss.sum()
        
        return loss

    def update_group_weights(self, group_loss):
        group_weights = self.group_weights
        group_weights = group_weights * torch.exp(self.group_weight_lr * group_loss)
        group_weights = group_weights / group_weights.sum()
        self.group_weights.data = group_weights.data

class GroupDRO(BaseRobustTrain):
    """
    Group DRO (https://arxiv.org/abs/1911.08731)
    """
    def __init__(
        self,
        model: nn.Module,
        trainset: GroupLabeledDatasetWrapper,
        batch_size: int,
        optimizer: optim.Optimizer,
        num_epochs: int,
        device: torch.device = torch.device("cpu"),
        val_evaluator: Evaluator = None,
        verbose=False,
    ):
        """
        Initializes GroupDRO.

        :param model: The PyTorch model to be trained.
        :type model: nn.Module
        :param trainset: The training dataset containing group-labeled samples.
        :type trainset: GroupLabeledDatasetWrapper
        :param batch_size: The batch size for training.
        :type batch_size: int
        :param optimizer: The optimizer used for training.
        :type optimizer: optim.Optimizer
        :param num_epochs: The number of training epochs.
        :type num_epochs: int
        :param device: The device to be used for training (default: CPU).
        :type device: torch.device
        :param verbose: Whether to print training progress (default: False).
        :type verbose: bool
        """

        seed_randomness(torch_module=torch, random_module=random, numpy_module=np)
    
        super().__init__(val_evaluator=val_evaluator, verbose=verbose)
    
        assert batch_size >= len(trainset.group_partition), "batch_size must be >= number of groups (Group DRO requires at least 1 example from each group)"

        def forward_pass(self, batch):
            inputs, labels, groups = batch
            inputs, labels, groups = inputs.to(self.device), labels.to(self.device), groups.to(self.device)
            outputs = self.model(inputs)
            loss = self.criterion(outputs, labels, groups)
            return loss, outputs, labels
        
        self.num_epochs = num_epochs
        self.group_partition = trainset.group_partition
        self.group_weighted_loss = GroupWeightedLoss(criterion=nn.CrossEntropyLoss(reduction="none"), num_groups=len(self.group_partition), device=device)
        self.trainer = Trainer(
            trainset=trainset,
            model=model,
            batch_size=batch_size,
            optimizer=optimizer,
            criterion=self.group_weighted_loss,
            forward_pass=forward_pass,
            sampler=CustomIndicesSampler(indices=[]),
            verbose=verbose,
            device=device
        )

        max_group_len = max([len(self.group_partition[key]) for key in self.group_partition.keys()])
        self.base_indices = []
        self.sampling_weights = []
        for key in self.group_partition.keys():
            self.base_indices.extend(self.group_partition[key])
            self.sampling_weights.extend([max_group_len / len(self.group_partition[key])] * len(self.group_partition[key]))
        
    def train_epoch(self, epoch):
        """
        Trains the model for a single epoch with a group balanced batch (in expectation)

        :param epoch: The current epoch number.
        :type epoch: int
        """
        self.trainer.sampler.indices = random.choices(
            population=self.base_indices,
            weights=self.sampling_weights, 
            k=len(self.trainer.trainset)
        )
        self.trainer.train_epoch(epoch)



import random

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import Dataset

from spuco.evaluate import Evaluator
from spuco.robust_train import BaseRobustTrain
from spuco.utils import Trainer
from spuco.utils.random_seed import seed_randomness


class ERM(BaseRobustTrain):
    """
    Empirical Risk Minimization (ERM) Trainer.
    """
    def __init__(
        self,
        model: nn.Module,
        trainset: Dataset,
        batch_size: int,
        optimizer: optim.Optimizer,
        num_epochs: int,
        criterion=nn.CrossEntropyLoss(), 
        device: torch.device = torch.device("cpu"),
        lr_scheduler=None,
        max_grad_norm=None,
        val_evaluator: Evaluator = None,
        verbose=False
    ):
        """
        Initializes a ERM instance.

        :param model: The neural network model to train.
        :type model: nn.Module
        :param trainset: The trainset to use for training.
        :type trainset: Dataset
        :param batch_size: The batch size to use during training.
        :type batch_size: int
        :param optimizer: The optimizer to use during training.
        :type optimizer: torch.optim.Optimizer
        :param num_epochs: The number of epochs to train for.
        :type num_epochs: int
        :param criterion: The loss function to use. Default is nn.CrossEntropyLoss().
        :type criterion: nn.Module, optional
        :param device: The device to use for training. Default is torch.device("cpu").
        :type device: torch.device, optional
        :param verbose: If True, prints verbose training information. Default is False.
        :type verbose: bool, optional
        """
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        super().__init__(val_evaluator=val_evaluator, verbose=verbose)

        self.num_epochs = num_epochs
        self.trainer = Trainer(
            trainset=trainset,
            model=model,
            batch_size=batch_size,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
            max_grad_norm=max_grad_norm,
            criterion=criterion,
            verbose=verbose,
            device=device
        )



from .base_robust_train import BaseRobustTrain
from .erm import ERM
from .group_dro import GroupDRO, GroupWeightedLoss
from .custom_sample_erm import CustomSampleERM
from .class_balance_batch_erm import ClassBalanceBatchERM
from .group_balance_batch_erm import GroupBalanceBatchERM
from .cnc_train import CorrectNContrastTrain
from .upsample_erm import UpSampleERM
from .downsample_erm import DownSampleERM
from .spare_train import SpareTrain


import random as random

import numpy as np
import torch
from transformers import DistilBertModel

from spuco.utils.random_seed import seed_randomness

# Acknowledgement to
# https://github.com/p-lambda/wilds/blob/main/examples/models/bert/distilbert.py


class DistilBert(DistilBertModel):
    def __init__(self, config):
        """
        Initializes a DistilBERT model.

        :param config: Configuration for the DistilBERT model.
        :type config: transformers.configuration_distilbert.DistilBertConfig
        """
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)
        super().__init__(config)
        self.d_out = config.hidden_size

    def __call__(self, x):
        """
        Forward pass of the DistilBERT model.

        :param x: Input tensor.
        :type x: torch.Tensor

        :return: Output tensor.
        :rtype: torch.Tensor
        """
        input_ids = x[:, :, 0]
        attention_mask = x[:, :, 1]
        hidden_state = super().__call__(
            input_ids=input_ids,
            attention_mask=attention_mask,
        )[0]
        pooled_output = hidden_state[:, 0]
        return pooled_output



import random
from typing import Dict

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import Dataset

from spuco.evaluate import Evaluator
from spuco.robust_train import BaseRobustTrain
from spuco.utils import CustomIndicesSampler, Trainer
from spuco.utils.random_seed import seed_randomness


class GroupBalanceBatchERM(BaseRobustTrain):
    """
    GroupBalanceBatchERM class for training a model using group balanced sampling.
    """
    def __init__(
        self,
        model: nn.Module,
        trainset: Dataset,
        group_partition: Dict,
        batch_size: int,
        optimizer: optim.Optimizer,
        num_epochs: int,
        device: torch.device = torch.device("cpu"),
        val_evaluator: Evaluator = None,
        verbose=False
    ):
        """
        Initializes GroupBalanceBatchERM.

        :param model: The PyTorch model to be trained.
        :type model: nn.Module
        :param trainset: The training dataset.
        :type trainset: Dataset
        :param group_partition: A dictionary mapping group labels to the indices of examples belonging to each group.
        :type group_partition: Dict
        :param batch_size: The batch size for training.
        :type batch_size: int
        :param optimizer: The optimizer used for training.
        :type optimizer: optim.Optimizer
        :param num_epochs: The number of training epochs.
        :type num_epochs: int
        :param device: The device to be used for training (default: CPU).
        :type device: torch.device
        :param verbose: Whether to print training progress (default: False).
        :type verbose: bool
        """
        
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)

        super().__init__(val_evaluator=val_evaluator, verbose=verbose)
        
        assert batch_size >= len(trainset.group_partition), "batch_size must be >= number of groups (Group DRO requires at least 1 example from each group)"
        
        self.num_epochs = num_epochs
        self.group_partition = group_partition
        self.trainer = Trainer(
            trainset=trainset,
            model=model,
            batch_size=batch_size,
            optimizer=optimizer,
            sampler=CustomIndicesSampler(indices=[]),
            verbose=verbose,
            device=device
        )

        max_group_len = max([len(self.group_partition[key]) for key in self.group_partition.keys()])
        self.base_indices = []
        self.sampling_weights = []
        for key in self.group_partition.keys():
            self.base_indices.extend(self.group_partition[key])
            self.sampling_weights.extend([max_group_len / len(self.group_partition[key])] * len(self.group_partition[key]))
        
    def train_epoch(self, epoch: int):
        """
        Trains the model for a single epoch with a group balanced batch (in expectation)

        :param epoch: The current epoch number.
        :type epoch: int
        """
        self.trainer.sampler.indices = random.choices(
            population=self.base_indices,
            weights=self.sampling_weights, 
            k=len(self.trainer.trainset)
        )
        self.trainer.train_epoch(epoch)



import random

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
from spuco.datasets import BaseSpuCoCompatibleDataset, GroupLabeledDatasetWrapper
from spuco.models import SpuCoModel
from spuco.utils.random_seed import seed_randomness
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from typing import List, Optional, Dict
import random


class DFR():
    def __init__(
        self,
        group_labeled_set: GroupLabeledDatasetWrapper,
        model: SpuCoModel,
        n_lin_models: int = 20, 
        labeled_valset_size: float = 0.5,
        C_range: List[float] = [1., 0.7, 0.3, 0.1, 0.07, 0.03, 0.01],
        class_weight_options: Optional[List[Dict]] = None,
        validation_set: Optional[GroupLabeledDatasetWrapper] = None,
        device: torch.device = torch.device("cpu"),
        verbose: bool = False,
        data_for_scaler: Optional[Dataset] = None
    ):
        """
        Initializes the DFR object.

        :param group_labeled_set: The group-labeled dataset.
        :type group_labeled_set: Dataset
        :param model: The base model.
        :type model: SpucoModel
        :param n_lin_models: Number of linear models to average.
        :type trainset: int
        :param labeled_valset_size: The ratio of the labeled data to be used for validation if no validation set is given. 
        :type labeled_valset_size: float
        :param C_range: Options of C, which is the inverse of l1 regularization strength as in sklearn.
        :type C_range: list
        :param class_weight_options: options for class weight.
        :type class_weight_options: list
        :param validation_set: data used for hyperparameter selection. If not provided, half of the group labeled data will be used.
        :type validation_set: GroupLabeledDatasetWrapper
        :param data_for_scaler: Data used for fitting the sklearn scaler. If not provided, group labeled data will be used.
        :type data_for_scaler: Dataset
        """
          
        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        self.group_labeled_set = group_labeled_set
        self.model = model 
        self.C_range = C_range
        self.n_lin_models = n_lin_models
        self.labeled_valset_size = labeled_valset_size
        self.device = device 
        self.verbose = verbose
        self.data_for_scaler = data_for_scaler
        self.scaler = None
        self.preprocess = self.data_for_scaler is not None
        self.class_weight_options = class_weight_options
        self.validation_set = validation_set
    

    def train_single_model(self, C, X_train, y_train, g_train, class_weight):
        """
        Trains a single model.

        :param C: Regularization parameter C for the SVM model.
        :type C: float

        :param X_train: Training features.
        :type X_train: numpy.ndarray

        :param y_train: Training labels.
        :type y_train: numpy.ndarray

        :param g_train: Training group labels.
        :type g_train: numpy.ndarray

        :param class_weight: Weight associated with each class.
        :type class_weight: dict or 'balanced', optional
        """
        group_partition = []
        for g in range(np.max(g_train)+1):
            group_partition.append(np.where(g_train==g)[0])
        min_size = np.min([len(g) for g in group_partition])
        X_train_balanced = []
        y_train_balanced = []
        for g in group_partition:
            indices = np.random.choice(g, size=min_size, replace=False)
            X_train_balanced.append(X_train[indices])
            y_train_balanced.append(y_train[indices])
        X_train_balanced = np.concatenate(X_train_balanced)
        y_train_balanced = np.concatenate(y_train_balanced)

        logreg = LogisticRegression(penalty='l1', C=C, solver="liblinear", class_weight=class_weight)
        logreg.fit(X_train_balanced, y_train_balanced)
        
        return logreg.coef_, logreg.intercept_

    def train_multiple_model(self, C, X_labeled_train, y_labeled_train, g_labeled_train, class_weight):
        """
        Trains the DFR model.

        :param C: Regularization parameter C for the SVM model.
        :type C: float

        :param X_labeled_train: Labeled training features.
        :type X_labeled_train: numpy.ndarray

        :param y_labeled_train: Labeled training labels.
        :type y_labeled_train: numpy.ndarray

        :param g_labeled_train: Labeled training group labels.
        :type g_labeled_train: numpy.ndarray

        :param class_weight: Weight associated with each class.
        :type class_weight: dict or 'balanced', optional
        """
        coefs, intercepts = [], []
        for i in range(self.n_lin_models):
            coef_, intercept_ = self.train_single_model(C, X_labeled_train, y_labeled_train, g_labeled_train, class_weight)
            coefs.append(coef_)
            intercepts.append(intercept_)
        return np.mean(coefs, axis=0), np.mean(intercepts, axis=0)
    

    def hyperparam_selection(self, X_labeled_train, y_labeled_train, g_labeled_train, X_labeled_val, y_labeled_val, g_labeled_val):
        """
        Performs hyperparameter selection for the DFR model.

        :param X_labeled_train: Labeled training features.
        :type X_labeled_train: numpy.ndarray

        :param y_labeled_train: Labeled training labels.
        :type y_labeled_train: numpy.ndarray

        :param g_labeled_train: Labeled training group labels.
        :type g_labeled_train: numpy.ndarray

        :param X_labeled_val: Labeled validation features.
        :type X_labeled_val: numpy.ndarray

        :param y_labeled_val: Labeled validation labels.
        :type y_labeled_val: numpy.ndarray

        :param g_labeled_val: Labeled validation group labels.
        :type g_labeled_val: numpy.ndarray
        """
        best_wg_acc = -1
        if self.verbose:
            print('Searching for best hyperparameters ...')

        for C in self.C_range:
            for class_weight in self.class_weight_options:
                coef, intercept = self.train_single_model(C, X_labeled_train, y_labeled_train, g_labeled_train, class_weight)
                # coef, intercept = self.train_multiple_model(C, X_labeled_train, y_labeled_train, g_labeled_train)
                wg_acc = self.evaluate_worstgroup_acc(C, coef, intercept, X_labeled_val, y_labeled_val, g_labeled_val)
                if wg_acc > best_wg_acc:
                    best_wg_acc = wg_acc
                    self.best_C = C
                    self.best_class_weight = class_weight
                if self.verbose:
                    print('C {} | class weight {}: Val Worst-Group Accuracy: {}'.format(C, class_weight, wg_acc))
                    print('Best C {} | class weight {}: Best Val Worst-Group Accuracy: {}'.format(self.best_C, self.best_class_weight, best_wg_acc))
    
    def train(self):
        """
        Retrain last layer
        """
        if self.verbose:
            print('Encoding data ...')
        X_labeled, y_labeled, g_labeled = self.encode_dataset(self.group_labeled_set)
        X_labeled = X_labeled.detach().cpu().numpy()
        y_labeled = y_labeled.detach().cpu().numpy()
        g_labeled = g_labeled.detach().cpu().numpy()

        # Standardize features
        if self.preprocess:
            self.scaler = StandardScaler()
            if self.data_for_scaler:
                X_scaler, _ = self.encode_dataset(self.data_for_scaler)
                X_scaler = X_scaler.detach().cpu().numpy()
                self.scaler.fit(X_scaler)
            else:
                self.scaler = StandardScaler()
                self.scaler.fit(X_labeled)
            X_labeled = self.scaler.transform(X_labeled)
        
        # If validation set is not provided, split labeled data into training and validation data
        # Otherwise, use the given validation set 
        if self.validation_set is None:
            n_labeled = X_labeled.shape[0]
            ids = {i for i in range(n_labeled)}
            ids_val = set(random.sample(ids, int(self.labeled_valset_size*n_labeled)))
            ids_train = ids - ids_val
            ids_val = np.array(list(ids_val))
            ids_train = np.array(list(ids_train))

            X_labeled_train, y_labeled_train, g_labeled_train = X_labeled[ids_train], y_labeled[ids_train], g_labeled[ids_train]
            X_labeled_val, y_labeled_val, g_labeled_val = X_labeled[ids_val], y_labeled[ids_val], g_labeled[ids_val]
        else:
            X_labeled_train, y_labeled_train, g_labeled_train = X_labeled, y_labeled, g_labeled
            X_labeled_val, y_labeled_val, g_labeled_val = self.encode_dataset(self.validation_set)
            X_labeled_val = X_labeled_val.detach().cpu().numpy()
            y_labeled_val = y_labeled_val.detach().cpu().numpy()
            g_labeled_val = g_labeled_val.detach().cpu().numpy()
            if self.preprocess:
                X_labeled_val = self.scaler.transform(X_labeled_val)
        
        if self.class_weight_options is None:
            n_class = np.max(y_labeled_val) + 1
            self.class_weight_options = [{c: 1 for c in range(n_class)}]
        
        self.hyperparam_selection(X_labeled_train, y_labeled_train, g_labeled_train, X_labeled_val, y_labeled_val, g_labeled_val)
        coef, intercept = self.train_multiple_model(self.best_C, X_labeled, y_labeled, g_labeled, self.best_class_weight)
        self.linear_model = (self.best_C, coef, intercept, self.scaler)


    def evaluate_worstgroup_acc(self, C, coef, intercept, X_val, y_val, g_val):
        """
        Evaluates the worst-group accuracy for the DFR model.

        :param C: Regularization parameter C for the SVM model.
        :type C: float

        :param coef: Coefficients of the linear SVM model.
        :type coef: numpy.ndarray

        :param intercept: Intercept of the linear SVM model.
        :type intercept: numpy.ndarray

        :param X_val: Validation features.
        :type X_val: numpy.ndarray

        :param y_val: Validation labels.
        :type y_val: numpy.ndarray

        :param g_val: Validation group labels.
        :type g_val: numpy.ndarray
        """
        logreg = LogisticRegression(penalty='l1', C=C, solver="liblinear")
        n_classes = np.max(y_val) + 1
        # the fit is only needed to set up logreg
        logreg.fit(X_val[:n_classes], np.arange(n_classes))
        
        logreg.coef_ = coef
        logreg.intercept_ = intercept
        
        preds = logreg.predict(X_val)
        n_groups = np.max(g_val) + 1
        accs = [(preds == y_val)[g_val == g].mean() for g in range(n_groups)]
        return np.min(accs)


    def encode_dataset(self, dataset):
        """
        Encodes the training set using the DFR model.

        :param dataset: The training dataset.
        :type dataset: torch.utils.data.Dataset

        :return: The encoded features and labels of the training set.
        :rtype: Tuple[torch.Tensor, torch.Tensor]
        """

        labeled = type(dataset) == GroupLabeledDatasetWrapper

        X_train = []
        y_train = []
        if labeled:
            g_train = []

        trainloader = DataLoader(
            dataset=dataset, 
            batch_size=100,
            shuffle=False,
            num_workers=4, 
            pin_memory=True
        )

        self.model.eval()
        with torch.no_grad():
            for batch in trainloader:
                if labeled:
                    inputs, labels, groups = batch
                    inputs, labels, groups = inputs.to(self.device), labels.to(self.device), groups.to(self.device)
                    g_train.append(groups)
                else:
                    inputs, labels = batch
                    inputs, labels = inputs.to(self.device), labels.to(self.device)
                X_train.append(self.model.backbone(inputs))
                y_train.append(labels)
                    
            if labeled:
                return torch.cat(X_train), torch.cat(y_train), torch.cat(g_train)
            else:
                return torch.cat(X_train), torch.cat(y_train)



import random
from typing import Dict, List, Tuple

import numpy as np
import torch

from spuco.group_inference import BaseGroupInference
from spuco.utils.random_seed import seed_randomness


class JTTInference(BaseGroupInference):
    """
    Just Train Twice Inference: https://arxiv.org/abs/2107.09044
    """
    def __init__(
        self,
        predictions: List[int],
        class_labels: List[int],
    ):
        """
        Initializes JTTInference.

        :param predictions: List of predicted labels.
        :param class_labels: List of true class labels.
        """

        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)

        super().__init__()
        self.predictions = predictions
        self.class_labels = class_labels

    def infer_groups(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Infers group partitions based on predictions and class labels.

        :return: Dictionary mapping group tuples to indices of examples belonging to each group.
        """
        group_partition = {(0, 0): [], (0, 1): []} 

        for i, label in enumerate(self.class_labels):
            if label == self.predictions[i]:
                group_partition[(0, 0)].append(i)
            else:
                group_partition[(0, 1)].append(i)

        return group_partition




from torch.utils.data import Dataset 
from typing import List

class SpuriousTargetDatasetWrapper(Dataset):
    """
    Wrapper class that takes a Dataset and the spurious labels of the data
    and returns a dataset where the labels are the spurious labels. 
    """
    def __init__(
        self,
        dataset: Dataset,
        spurious_labels: List[int]
    ):
        """
        Initialize an instance of SpuriousTargetDatasetWrapper.

        :param dataset: The original dataset.
        :type dataset: Dataset
        :param spurious_labels: The spurious labels corresponding to the data.
        :type spurious_labels: List[int]
        """
        self.dataset = dataset
        self.spurious_labels = spurious_labels

    def __getitem__(self, index):
        """
        Get an item from the dataset.

        :param index: The index of the item to retrieve.
        :type index: int
        :return: A tuple containing the input data and the spurious label.
        :rtype: Tuple[Any, int]
        """
        return (self.dataset.__getitem__(index)[0], self.spurious_labels[index])
    
    def __len__(self):
        """
        Get the length of the dataset.

        :return: The length of the dataset.
        :rtype: int
        """
        return len(self.dataset)
        


import heapq
import math

import numpy as np
from tqdm import tqdm


class FacilityLocation:
    def __init__(self, D, V, alpha=1.):
        '''
        Args
        - D: np.array, shape [N, N], similarity matrix
        - V: list of int, indices of columns of D
        - alpha: float
        '''
        self.D = D
        self.curVal = 0
        self.curMax = np.zeros(len(D))
        self.gains = []
        self.alpha = alpha
        self.f_norm = self.alpha / self.f_norm(V)
        self.norm = 1. / self.inc(V, [])

    def f_norm(self, sset):
        return self.D[:, sset].max(axis=1).sum()

    def inc(self, sset, ndx):
        if len(sset) > 0:
            if not ndx:  # normalization
                return math.log(1 + self.alpha * 1)
            return self.norm * math.log(1 + self.f_norm * np.maximum(self.curMax, self.D[:, ndx]).sum()) - self.curVal
        else:
            return self.norm * math.log(1 + self.f_norm * self.D[:, ndx].sum()) - self.curVal

    def add(self, sset, ndx):
        cur_old = self.curVal
        if len(sset + [ndx]) > 1:
            self.curMax = np.maximum(self.curMax, self.D[:, ndx])
        else:
            self.curMax = self.D[:, ndx]
        self.curVal = self.norm * math.log(1 + self.f_norm * self.curMax.sum())
        self.gains.extend([self.curVal - cur_old])
        return self.curVal


def _heappush_max(heap, item):
    heap.append(item)
    heapq._siftdown_max(heap, 0, len(heap)-1)


def _heappop_max(heap):
    """Maxheap version of a heappop."""
    lastelt = heap.pop()  # raises appropriate IndexError if heap is empty
    if heap:
        returnitem = heap[0]
        heap[0] = lastelt
        heapq._siftup_max(heap, 0)
        return returnitem
    return lastelt


def lazy_greedy(F, V, B, verbose=False):
    """
    Args
    - F: FacilityMaximization Object
    - V: list of indices of columns of Similarity Matrix
    - B: Budget of subset (int)
    """
    curVal = 0
    sset = []
    vals = []

    order = []
    heapq._heapify_max(order)
    [_heappush_max(order, (F.inc(sset, index), index)) for index in V]

    if verbose:
        print("Submodular Maximization: Selecting subset...")
    with tqdm(total=B, disable=not verbose) as pbar:
        while order and len(sset) < B:
            el = _heappop_max(order)
            improv = F.inc(sset, el[1])

            if improv >= 0:
                if not order:
                    curVal = F.add(sset, el[1])
                    sset.append(el[1])
                    vals.append(curVal)
                    pbar.update(1)
                else:
                    top = _heappop_max(order)
                    if improv >= top[0]:
                        curVal = F.add(sset, el[1])
                        sset.append(el[1])
                        vals.append(curVal)
                        pbar.update(1)
                    else:
                        _heappush_max(order, (improv, el[1]))
                    _heappush_max(order, top)
    return sset, vals


from typing import Dict, List, Optional, Tuple

import torch
from torch.utils.data import Dataset

class GroupLabeledDatasetWrapper(Dataset):
    def __init__(
        self, 
        dataset: Dataset,
        group_partition: Dict[Tuple[int, int], int],
        subset_indices: Optional[List[int]] = None
    ):
        """
        Initializes a GroupLabeledDataset.

        :param dataset: The underlying dataset.
        :type dataset: torch.utils.data.Dataset

        :param group_partition: The group partition dictionary mapping indices to group labels.
        :type group_partition: Dict[Tuple[int, int], int]

        :param subset_indices: Optional list of subset indices to consider from the dataset. Defaults to None.
        :type subset_indices: Optional[List[int]]
        """
        self.dataset = dataset
        self.group = torch.zeros(len(self.dataset))
        self.group_partition = group_partition
        
        group_idx = 0
        for key in sorted(group_partition.keys()):
            self.group[group_partition[key]] = group_idx
            group_idx += 1 
        self.num_groups = len(group_partition.keys())
        self.group = self.group.long().tolist()
        
        # Subset if needed
        self.indices = range(len(dataset))
        if subset_indices is not None:
            self.indices = subset_indices
        
    def __getitem__(self, index: int):
        """
        Retrieves an item from the dataset.

        :param index: The index of the item.
        :type index: int
        :return: The item at the given index.
        """
        index = self.indices[index]
        source_tuple = self.dataset.__getitem__(index)
        return (source_tuple[0], source_tuple[1], self.group[index])
    
    def __len__(self):
        """
        Returns the length of the dataset.

        :return: The length of the dataset.
        :rtype: int
        """
        return len(self.indices)


import random
from collections import Counter
from copy import deepcopy
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import DataLoader, Dataset, Subset
from tqdm import tqdm

from spuco.datasets import BaseSpuCoCompatibleDataset, SpuriousTargetDatasetWrapper
from spuco.group_inference import BaseGroupInference
from spuco.utils import Trainer
from spuco.utils.random_seed import seed_randomness


class SSA(BaseGroupInference):
    """
    Spread Spurious Attribute: https://arxiv.org/abs/2204.02070
    """
    def __init__(
        self, 
        spurious_unlabeled_dataset: BaseSpuCoCompatibleDataset, 
        spurious_labeled_dataset: SpuriousTargetDatasetWrapper,
        model: nn.Module, 
        num_iters: int,
        labeled_valset_size: float = 0.5,
        tau_g_min: float = 0.95,
        lr: float = 1e-2,
        weight_decay: float = 5e-4,
        batch_size: int = 64,
        num_splits: int = 3, 
        device: torch.device = torch.device("cpu"),
        verbose: bool = False
    ):
        """
        Initializes SSA.

        :param spurious_unlabeled_dataset: The dataset containing spurious-labeled unlabeled samples.
        :type spurious_unlabeled_dataset: Dataset
        :param spurious_labeled_dataset: The dataset containing spurious-labeled labeled samples.
        :type spurious_labeled_dataset: SpuriousTargetDatasetWrapper
        :param model: The PyTorch model to be used.
        :type model: nn.Module
        :param labeled_valset_size: The size of the labeled validation set as a fraction of the total labeled dataset size.
        :type labeled_valset_size: float
        :param num_iters: The number of iterations for SSA.
        :type num_iters: int
        :param tau_g_min: The minimum value of the spurious correlation threshold tau_g.
        :type tau_g_min: float
        :param lr: The learning rate for optimization (default: 1e-2).
        :type lr: float
        :param weight_decay: The weight decay for optimization (default: 5e-4).
        :type weight_decay: float
        :param batch_size: The batch size for training (default: 64).
        :type batch_size: int
        :param num_splits: The number of splits for cross-validation (default: 3).
        :type num_splits: int
        :param device: The device to be used for training (default: CPU).
        :type device: torch.device
        :param verbose: Whether to print training progress (default: False).
        :type verbose: bool
        """

        seed_randomness(torch_module=torch, numpy_module=np, random_module=random)
        super().__init__()

        self.spurious_unlabeled_dataset = spurious_unlabeled_dataset
        self.spurious_labeled_dataset = spurious_labeled_dataset
        self.model = model 
        self.lr = lr
        self.weight_decay = weight_decay
        self.batch_size = batch_size 
        self.num_iters = num_iters
        self.num_splits = num_splits 
        self.tau_g_min = tau_g_min
        self.device = device 
        self.verbose = verbose 
        
        # Create unlabeled data splits
        self.unlabeled_data_splits = []
        indices = np.array(range(len(self.spurious_unlabeled_dataset)))
        indices = indices[torch.randperm(len(indices)).numpy()]
        self.unlabeled_data_splits = np.array_split(indices, self.num_splits)
        
        # Create labeled data splits
        indices = np.array(range(len(self.spurious_labeled_dataset)))
        indices = indices[torch.randperm(len(indices)).numpy()]
        self.labeled_val_indices = indices[:int(len(indices) * labeled_valset_size)]
        self.labeled_train_indices = indices[int(len(indices) * labeled_valset_size):]

        # Get class labels for unlabeled set
        self.class_labels = spurious_unlabeled_dataset.labels

        # Determine g_min 
        group_counter = Counter(np.array(spurious_labeled_dataset.spurious_labels)[self.labeled_val_indices])
        self.g_min = group_counter.most_common()[-1][0]

    def infer_groups(self) -> Dict[Tuple[int, int], List[int]]:
        """
        Infer groups based on spurious attribute predictions.

        :return: A dictionary mapping group labels to the indices of examples belonging to each group.
        :rtype: Dict[Tuple[int, int], List[int]]
        """
        # Train SSA models to get spurious target predictions
        spurious_labels = np.zeros(len(self.spurious_unlabeled_dataset))
        for split_num in range(self.num_splits):
            best_ssa_model = self.train_ssa(split_num)
            curr_split_labels = self.label_split(split_num, best_ssa_model)
            spurious_labels[self.unlabeled_data_splits[split_num]] = curr_split_labels
        spurious_labels = spurious_labels.astype(int).tolist()

        # Convert class and spurious labels into group partition
        group_partition = {}
        for i in range(len(self.spurious_unlabeled_dataset)):
            spurious_label = (self.class_labels[i], spurious_labels[i])
            if spurious_label not in group_partition:
                group_partition[spurious_label] = []
            group_partition[spurious_label].append(i)
        return group_partition
    
    def train_ssa(self, split_num: int) -> nn.Module:
        """
        Train an SSA model for the specified split.

        :param split_num: The index of the split to train the SSA model for.
        :type split_num: int
        :return: The trained SSA model.
        :rtype: nn.Module
        """
        trainer = SSATrainer(self, split_num)
        trainer.train()
        if self.verbose:
            print("Best Validation Accuracy", trainer.best_acc)
        return trainer.best_model
    
    def label_split(self, split_num: int, best_ssa_model: nn.Module) -> np.array:
        """
        Label the specified split using the best SSA model.

        :param split_num: The index of the split to label.
        :type split_num: int
        :param best_ssa_model: The best SSA model obtained during training.
        :type best_ssa_model: nn.Module
        :return: An array of spurious labels for the split.
        :rtype: np.array
        """
        split_dataloader = DataLoader(
            dataset=Subset(
                dataset=self.spurious_unlabeled_dataset,
                indices=self.unlabeled_data_splits[split_num]
            ),
            batch_size=self.batch_size, 
            shuffle=False,
            num_workers=4, 
            pin_memory=True
        )

        with torch.no_grad():
            preds = []
            for X, _ in split_dataloader:
                X = X.to(self.device)
                preds.append(torch.argmax(best_ssa_model(X), dim=-1))
            preds = torch.cat(preds, dim=0).cpu().numpy()
            return preds

class SSATrainer:
    def __init__(
            self,
            ssa: SSA,
            split_num: int,
    ):
        """
        Initializes SSATrainer.
        :param ssa: The SSA object containing the data, model etc. for training.
        :type ssa: SSA
        :param split_num: The index of the split to train the SSA model for.
        :type split_num: int
        """
        self.ssa = ssa
        self.tau_g_min = self.ssa.tau_g_min
        self.device = self.ssa.device
        self.best_model = None
        self.model = deepcopy(self.ssa.model).to(self.device)
        self.optimizer = optim.SGD(self.model.parameters(), lr=self.ssa.lr, weight_decay=self.ssa.weight_decay)
        self.split_num = split_num
        self.num_iters = self.ssa.num_iters
        self.verbose = ssa.verbose
        
        # Create member variable for best model, accuracy
        self.best_model = None
        self.best_acc = -1.

        self.unlabeled_trainloader = DataLoader(
            dataset=Subset(
                dataset=self.ssa.spurious_unlabeled_dataset,
                indices=[i for split_num in range(self.ssa.num_splits) if split_num != self.split_num for i in self.ssa.unlabeled_data_splits[split_num]]
            ),
            batch_size=self.ssa.batch_size, 
            shuffle=True,
            num_workers=4, 
            pin_memory=True
        )

        self.labeled_trainloader = DataLoader(
            dataset=Subset(
                dataset=self.ssa.spurious_labeled_dataset,
                indices=self.ssa.labeled_train_indices
            ),
            batch_size=self.ssa.batch_size, 
            shuffle=True,
            num_workers=4, 
            pin_memory=True
        )
        self.valloader = DataLoader(
            dataset=Subset(
                dataset=self.ssa.spurious_labeled_dataset,
                indices=self.ssa.labeled_val_indices
            ),
            batch_size=self.ssa.batch_size, 
            shuffle=False,
            num_workers=4, 
            pin_memory=True
        )

        self.cross_entropy = nn.CrossEntropyLoss()

    def train(self) -> None:
        """
        Trains model targetting spurious attribute for given split.
        """
        with tqdm(range(self.num_iters), total=self.num_iters, disable=not self.verbose) as pbar:
            pbar.set_description(f"Training SSA Model {self.split_num}")
            unlabeled_train_iter = iter(self.unlabeled_trainloader)
            labeled_train_iter = iter(self.labeled_trainloader)
            for _ in pbar:
                self.model.train()
                try:            
                    unlabeled_train_batch = next(unlabeled_train_iter)
                except StopIteration:
                    unlabeled_train_iter = iter(self.unlabeled_trainloader)
                    unlabeled_train_batch = next(unlabeled_train_iter)

                try:
                    labeled_train_batch = next(labeled_train_iter)
                except StopIteration:
                    labeled_train_iter = iter(self.labeled_trainloader)
                    labeled_train_batch = next(labeled_train_iter)

                # Compute loss
                loss = self.train_step(
                    unlabeled_train_batch=unlabeled_train_batch,
                    labeled_train_batch=labeled_train_batch
                )   
                loss.backward()
                self.optimizer.step()

                # Validation
                val_acc = self.validate()
                pbar.set_postfix(loss=loss.item(), val_acc=val_acc)
                
                # If perfect validation accuracy, no point training on since we pick best model based on this.
                if val_acc == 100:
                    return
    
    
    def train_step(self, unlabeled_train_batch, labeled_train_batch):
        """
        Trains a single step of SSA for a given batch of unlabeled and labeled data.

        :param unlabeled_train_batch: The batch of unlabeled training data.
        :type unlabeled_train_batch: torch.Tensor
        :param labeled_train_batch: The batch of labeled training data.
        :type labeled_train_batch: torch.Tensor
        """
        ###########################
        # Compute supervised loss #
        ###########################
        X, labels = labeled_train_batch
        X, labels = X.to(self.device), labels.to(self.device)
        outputs = self.model(X)
        supervised_loss = self.cross_entropy(outputs, labels)
        group_counter = Counter(labels.tolist())
        g_min_count = group_counter[self.ssa.g_min]

        #############################
        # Compute unsupervised loss #
        #############################
        # Load data
        X, y = unlabeled_train_batch
        X = X.to(self.device)

        # Get pseudo-labels
        outputs = self.model(X)
        pseudo_labels = torch.argmax(outputs, dim=-1)

        # Get indices of g_min outputs that are >= tau_g_min (threshold: hyperparameter)
        g_min_indices = torch.nonzero(pseudo_labels == self.ssa.g_min, as_tuple=True)[0]
        g_min_indices = g_min_indices[torch.nonzero(torch.max(outputs[g_min_indices], dim=-1).values >= self.tau_g_min, as_tuple=True)[0]]

        # Get all unsupervised indices for loss 
        unsup_indices = g_min_indices.cpu().tolist()
        for label, count in group_counter.most_common():
            if label == self.ssa.g_min:
                continue 
            # Determine which indices in current batch belong to this group 
            group_indices = torch.nonzero(pseudo_labels == label, as_tuple=True)[0]
            # Get prob of current label being the spurious attribute according to model
            group_outputs = outputs[group_indices, label]
            # Calculate the number of examples we should choose from this group
            k = min(max(g_min_count + len(g_min_indices) - count, 0), len(group_indices))
            # Get k indices with largest prob of group label being current label
            group_indices = group_indices[torch.topk(group_outputs, k=k).indices].cpu().tolist()
            unsup_indices.extend(group_indices)

        if len(unsup_indices) > 0:
            # Compute cross entropy with respect to pseudo-labels 
            unsupervised_loss = self.cross_entropy(outputs[unsup_indices], pseudo_labels[unsup_indices])
            return supervised_loss + unsupervised_loss
        else:
            return supervised_loss
    
    def validate(self):
        """
        Validates the SSA Model on the spurious attribute labeled validation set.
        """
        self.model.eval()
        with torch.no_grad():
            # Compute accuracy on validation
            outputs = []
            labels = []
            for X, label in self.valloader:
                X, label = X.to(self.device), label.to(self.device)
                outputs.append(self.model(X))
                labels.append(label)
            outputs = torch.cat(outputs, dim=0)
            labels = torch.cat(labels, dim=0)

            acc = Trainer.compute_accuracy(outputs, labels)

            # Update best model based on validation 
            if acc >= self.best_acc:
                self.best_acc = acc 
                self.best_model = deepcopy(self.model)
            return acc


import random

import numpy as np
import torch
import torch.nn as nn

from spuco.utils.random_seed import seed_randomness


class MLP(nn.Module):
    def __init__(
        self,
        input_dim: int,
    ):
        """
        Initializes the model.

        :param input_dim: Dimensionality of the input.
        :type input_dim: int
        """
        seed_randomness(random_module=random, torch_module=torch, numpy_module=np)
        super(MLP, self).__init__()
        self.input_dim = input_dim
        self.representation_dim = 256
        self.fc1 = nn.Linear(input_dim, 512)
        self.fc2 = nn.Linear(512, self.representation_dim)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=0.5)
        
    def forward(self, x):
        """
        Forward pass of the model.

        :param x: Input tensor.
        :type x: torch.Tensor

        :return: Output tensor.
        :rtype: torch.Tensor
        """
        x = x.view(-1, self.input_dim)
        x = self.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.relu(self.fc2(x))
        return x



import torch
from transformers import BertTokenizerFast, DistilBertTokenizerFast


def initialize_bert_transform(model, max_token_length):
    """
    Initializes the BERT transform for text data.

    :param model: The BERT model name.
    :type model: str
    :param max_token_length: The maximum token length for padding/truncation.
    :type max_token_length: int
    :return: The transform function.
    :rtype: Callable
    :raises ValueError: If the specified model is not recognized.
    """
    
    def get_bert_tokenizer(model):
        """
        Retrieves the appropriate BERT tokenizer based on the model name.

        :param model: The BERT model name.
        :type model: str
        :return: The BERT tokenizer.
        :rtype: PreTrainedTokenizerFast
        :raises ValueError: If the specified model is not recognized.
        """
        if model == "bert-base-uncased":
            return BertTokenizerFast.from_pretrained(model)
        elif model == "distilbert-base-uncased":
            return DistilBertTokenizerFast.from_pretrained(model)
        else:
            raise ValueError(f"Model: {model} not recognized.")

    tokenizer = get_bert_tokenizer(model)

    def transform(text):
        """
        Transforms the text data using BERT tokenizer.

        :param text: The input text.
        :type text: str
        :return: The transformed input as a PyTorch tensor.
        :rtype: torch.Tensor
        """
        tokens = tokenizer(
            text,
            padding="max_length",
            truncation=True,
            max_length=max_token_length,
            return_tensors="pt",
        )
        if model == "bert-base-uncased":
            x = torch.stack(
                (
                    tokens["input_ids"],
                    tokens["attention_mask"],
                    tokens["token_type_ids"],
                ),
                dim=2,
            )
        elif model == "distilbert-base-uncased":
            x = torch.stack((tokens["input_ids"], tokens["attention_mask"]), dim=2)
        x = torch.squeeze(x, dim=0)  # First shape dim is always 1
        return x

    return transform